import hashlib
import sys
import textwrap

from pytype.pyi import parser
from pytype.pyi import parser_test_base
from pytype.pytd import pytd
from pytype.tests import test_base
from pytype.tests import test_utils

import unittest


class ParseErrorTest(unittest.TestCase):

  def check(self, expected, *args, **kwargs):
    e = parser.ParseError(*args, **kwargs)
    self.assertMultiLineEqual(textwrap.dedent(expected).lstrip("\n"), str(e))

  def test_plain_error(self):
    self.check("""
        ParseError: my message""", "my message")

  def test_full_error(self):
    self.check("""
          File: "foo.py", line 123
            this is a test
                 ^
        ParseError: my message""", "my message", line=123, filename="foo.py",
               text="this is a test", column=6)

  def test_indented_text(self):
    self.check("""
          File: "foo.py", line 123
            this is a test
                 ^
        ParseError: my message""", "my message", line=123, filename="foo.py",
               text="          this is a test", column=16)

  def test_line_without_filename(self):
    self.check("""
          File: "None", line 1
        ParseError: my message""", "my message", line=1)

  def test_filename_without_line(self):
    self.check("""
          File: "foo.py", line None
        ParseError: my message""", "my message", filename="foo.py")

  def test_text_without_column(self):
    self.check("""
        ParseError: my message""", "my message", text="this is  a test")

  def test_column_without_text(self):
    self.check("        ParseError: my message", "my message", column=5)


class ParserTest(parser_test_base.ParserTestBase):

  def test_syntax_error(self):
    self.check_error("123", 1, "Unexpected expression")

  def test_illegal_character(self):
    self.check_error("^", 1, "invalid syntax")

  def test_invalid_indentation(self):
    self.check_error("""
      class Foo:
        x = ... # type: int
       y""", 3, "unindent does not match")

  @unittest.skip("New parser does not support this")
  def test_type_on_next_line(self):
    self.check("""
      a = ...
      # type: int""",
               """
      a: int""")

  def test_constant(self):
    self.check("x = ...", "x: Any", "from typing import Any")
    self.check("x: str")
    self.check("x = 0", "x: int")
    self.check("x = 0.0", "x: float")

  @unittest.skip("Not checking invalid literals")
  def test_invalid_constant(self):
    self.check_error("x = 123", 1,
                     "Only '0' allowed as int literal")
    self.check("x = 0.0", "x: float")
    self.check_error("x = 12.3", 1,
                     "Only '0.0' allowed as float literal")

  @test_utils.skipFromPy((3, 8), "No longer tagging strings 'unicode'")
  def test_string_constant_37(self):
    self.check("x = b''", "x: bytes")
    self.check("x = u''", "x: unicode")
    self.check('x = b""', "x: bytes")
    self.check('x = u""', "x: unicode")
    self.check("x = ''", "x: str")
    self.check('x = ""', "x: str")

  @test_utils.skipBeforePy((3, 8), "No longer tagging strings 'unicode'")
  def test_string_constant_38(self):
    self.check("x = b''", "x: bytes")
    self.check("x = u''", "x: str")
    self.check('x = b""', "x: bytes")
    self.check('x = u""', "x: str")
    self.check("x = ''", "x: str")
    self.check('x = ""', "x: str")

  @unittest.skip("We allow all strings.")
  def test_invalid_string_constant(self):
    self.check_error("x = b'x'", 1,
                     "Only '', b'', and u'' allowed as string literals")
    self.check_error("x = u'x'", 1,
                     "Only '', b'', and u'' allowed as string literals")
    self.check_error("x = 'x'", 1,
                     "Only '', b'', and u'' allowed as string literals")

  def test_constant_pep526(self):
    self.check("x : str", "x: str")
    self.check("x : str = ...", "x: str = ...")

  def test_alias_or_constant(self):
    self.check("x = True", "x: bool")
    self.check("x = False", "x: bool")
    self.check("x = Foo")
    self.check("""
      class A:
          x = True""", """
      class A:
          x: bool
    """)
    self.check("""
      class A:
          x = ...  # type: int
          y = x
          z = y""", """
      class A:
          x: int
          y: int
          z: int
    """)

  def test_method_aliases(self):
    self.check("""
      class A:
          def x(self) -> int: ...
          y = x
          z = y
          @classmethod
          def a(cls) -> str: ...
          b = a
          c = b""", """
      class A:
          def x(self) -> int: ...
          @classmethod
          def a(cls) -> str: ...
          def y(self) -> int: ...
          def z(self) -> int: ...
          @classmethod
          def b(cls) -> str: ...
          @classmethod
          def c(cls) -> str: ...
    """)

  def test_chained_assignment(self):
    self.check("""
      a = b = int
    """, """
      a = int
      b = int
    """)

  def test_multiple_assignment(self):
    self.check("""
      a, b = int, str
    """, """
      a = int
      b = str
    """)
    self.check("""
      (a, b) = (c, d) = int, str
    """, """
      a = int
      b = str
      c = int
      d = str
    """)

  def test_invalid_multiple_assignment(self):
    self.check_error("""
      a, b = int, str, bool
    """, 1, "Cannot unpack 2 values for multiple assignment")
    self.check_error("""
      a, b = int
    """, 1, "Cannot unpack 2 values for multiple assignment")

  def test_slots(self):
    self.check("""
      class A:
          __slots__ = ...  # type: tuple
    """, """
      class A: ...
    """)
    self.check("""
      class A:
          __slots__ = ["foo", "bar", "baz"]
    """)
    self.check("""
      class A:
          __slots__ = []
    """)
    self.check_error("""
      __slots__ = ["foo", "bar"]
    """, 1, "__slots__ only allowed on the class level")
    self.check_error("""
      class A:
          __slots__ = ["foo", "bar"]
          __slots__ = ["foo", "bar", "baz"]
    """, 1, "Duplicate __slots__ declaration")
    self.check_error("""
      class A:
          __slots__ = ["foo", ?]
    """, 2, "invalid syntax")
    self.check_error("""
      class A:
          __slots__ = int
    """, 2, "__slots__ must be a list of strings")

  def test_nested_class(self):
    self.check("""
      class A:
          class B: ...
    """)

  def test_nested_class_alias(self):
    self.check("""
      class A:
          class B: ...
          C = A.B
    """, """
      from typing import Type

      class A:
          class B: ...
          C: Type[A.B]
    """)

  def test_nested_class_module_alias(self):
    self.check("""
      class A:
          class B: ...
      C = A.B
    """, """
      from typing import Type

      C: Type[A.B]

      class A:
          class B: ...
    """)

  def test_conditional_nested_class(self):
    self.check("""
      if sys.version_info < (3, 5):
        class A:
          class B: ...
    """, "")

  def test_import(self):
    self.check("import foo.bar.baz")
    self.check("import a as b")
    self.check("from foo.bar import baz")
    self.check("from foo.bar import baz as abc")
    self.check("from typing import NamedTuple, TypeVar", "")
    self.check("from foo.bar import *")
    self.check_error("from foo import * as bar", 1, "invalid syntax")
    self.check("from foo import a, b",
               "from foo import a\nfrom foo import b")
    self.check("from foo import (a, b)",
               "from foo import a\nfrom foo import b")
    self.check("from foo import (a, b, )",
               "from foo import a\nfrom foo import b")

  def test_from_import(self):
    ast = self.check("from foo import c\nclass Bar(c.X): ...",
                     parser_test_base.IGNORE)
    base, = ast.Lookup("Bar").bases
    self.assertEqual(base, pytd.NamedType("foo.c.X"))

  def test_duplicate_names(self):
    self.check_error("""
      def foo() -> int: ...
      foo = ... # type: int""",
                     None,
                     "Duplicate top-level identifier(s): foo")
    self.check_error("""
      from x import foo
      def foo() -> int: ...""",
                     None,
                     "Duplicate top-level identifier(s): foo")
    self.check_error("""
      X = ... # type: int
      class X: ...""",
                     None,
                     "Duplicate top-level identifier(s): X")
    self.check_error("""
      X = ... # type: int
      X = TypeVar('X')""",
                     None,
                     "Duplicate top-level identifier(s): X")
    # A function is allowed to appear multiple times.
    self.check("""
      def foo(x: int) -> int: ...
      def foo(x: str) -> str: ...""",
               """
      from typing import overload

      @overload
      def foo(x: int) -> int: ...
      @overload
      def foo(x: str) -> str: ...""")
    # @overload decorators should be properly round-tripped.
    self.check("""
      @overload
      def foo(x: int) -> int: ...
      @overload
      def foo(x: str) -> str: ...""", """
      from typing import overload

      @overload
      def foo(x: int) -> int: ...
      @overload
      def foo(x: str) -> str: ...""")

  def test_type(self):
    self.check("x: str")
    self.check("x = ...  # type: (str)", "x: str")
    self.check("x: foo.bar.Baz", prologue="import foo.bar")
    self.check("x: nothing")

  @unittest.skip("TODO: add errors for these")
  def test_deprecated_type(self):
    self.check_error("x = ...  # type: int and str and float",
                     1, "invalid syntax")
    self.check_error("x = ...  # type: ?", 1, "invalid syntax")
    self.check("x = ...  # type: int or str or float",
               1, "invalid syntax")

  def test_empty_union_or_intersection_or_optional(self):
    self.check_error("def f(x: typing.Union): ...", 1,
                     "Missing options to typing.Union")
    self.check_error("def f(x: typing.Intersection): ...", 1,
                     "Missing options to typing.Intersection")
    self.check_error("def f(x: typing.Optional): ...", 1,
                     "Missing options to typing.Optional")

  def test_optional_extra_parameters(self):
    self.check_error("def f(x: typing.Optional[int, str]): ...", 1,
                     "Too many options to typing.Optional")

  def test_alias_lookup(self):
    self.check("""
      from somewhere import Foo
      x = ...  # type: Foo
      """, """
      import somewhere
      from somewhere import Foo

      x: somewhere.Foo""")

  def test_external_alias(self):
    self.check("""
      from somewhere import Foo

      class Bar:
          Baz = Foo
    """, """
      from somewhere import Foo
      from typing import Any

      class Bar:
          Baz: Any
    """)

  def test_same_named_alias(self):
    self.check("""
      import somewhere
      class Bar:
          Foo = somewhere.Foo
    """, """
      import somewhere
      from typing import Any

      class Bar:
          Foo: Any
    """)

  def test_type_params(self):
    ast = self.check("""
      from typing import TypeVar

      T = TypeVar('T')

      def func(x: T) -> T: ...""")
    # During parsing references to type paraemters are instances of NamedType.
    # They should be replaced by TypeParameter objects during post-processing.
    sig = ast.functions[0].signatures[0]
    self.assertIsInstance(sig.params[0].type, pytd.TypeParameter)
    self.assertIsInstance(sig.return_type, pytd.TypeParameter)

    # Check various illegal TypeVar arguments.
    self.check_error("T = TypeVar()", 1, "Missing arguments to TypeVar")
    self.check_error("T = TypeVar(*args)", 1, "Bad arguments to TypeVar")
    self.check_error("T = TypeVar(...)", 1, "Bad arguments to TypeVar")
    self.check_error("T = TypeVar('Q')", 1,
                     "TypeVar name needs to be 'Q' (not 'T')")
    self.check_error("T = TypeVar('T', covariant=True, int, float)", 1,
                     "positional argument follows keyword argument")
    self.check_error("T = TypeVar('T', rumpelstiltskin=True)", 1,
                     "Unrecognized keyword")

  def test_type_param_arguments(self):
    self.check("""
      from typing import List, TypeVar

      T = TypeVar('T', List[int], List[str])""")
    self.check("""
      from typing import List, TypeVar

      T = TypeVar('T', bound=List[str])""")
    # 'covariant' and 'contravariant' are ignored for now.
    self.check("""
      from typing import TypeVar

      T = TypeVar('T', str, unicode, covariant=True)""", """
      from typing import TypeVar

      T = TypeVar('T', str, unicode)""")
    self.check("""
      import other_mod
      from typing import TypeVar

      T = TypeVar('T', other_mod.A, other_mod.B)""")

  def test_typing_typevar(self):
    self.check("""
      import typing
      T = typing.TypeVar('T')
    """, """
      import typing
      from typing import TypeVar

      T = TypeVar('T')
    """)

  def test_error_formatting(self):
    src = """
      class Foo:
        this is not valid"""
    with self.assertRaises(parser.ParseError) as e:
      parser.parse_string(textwrap.dedent(src).lstrip(), filename="foo.py",
                          options=self.options)
    self.assertMultiLineEqual(textwrap.dedent("""
        File: "foo.py", line 2
          this is not valid
         ^
      ParseError: Unexpected expression
    """).strip("\n"), str(e.exception))

  def test_pep484_translations(self):
    ast = self.check("""
      x: None""")
    self.assertEqual(pytd.NamedType("NoneType"), ast.constants[0].type)

  def test_module_name(self):
    ast = self.check("x = ...  # type: int",
                     "foo.x: int",
                     name="foo")
    self.assertEqual("foo", ast.name)

  def test_no_module_name(self):
    # If the name is not specified, it is a digest of the source.
    src = ""
    ast = self.check(src)
    self.assertEqual(hashlib.md5(src.encode()).hexdigest(), ast.name)
    src = "x: int"
    ast = self.check(src)
    self.assertEqual(hashlib.md5(src.encode()).hexdigest(), ast.name)

  def test_pep84_aliasing(self):
    # This should not be done for the typing module itself.
    self.check("x = ... # type: Hashable",
               "typing.x: Hashable",
               name="typing")

  def test_module_class_clash(self):
    ast = parser.parse_string(textwrap.dedent("""
      from bar import X
      class bar:
        X = ... # type: Any
      y = bar.X.Baz
      z = X.Baz
    """), name="foo", options=self.options)
    self.assertEqual("foo.bar.X.Baz", ast.Lookup("foo.y").type.name)
    self.assertEqual("bar.X.Baz", ast.Lookup("foo.z").type.name)

  def test_trailing_list_comma(self):
    self.check("""
      from typing import Any, Callable

      x: Callable[
        [
          int,
          int,
        ],
        Any,
      ]
    """, """
      from typing import Any, Callable

      x: Callable[[int, int], Any]
    """)

  def test_all(self):
    self.check("""
      __all__ = ['a']
    """, """
      from typing import List

      __all__: List[str] = ...
    """)

  def test_invalid_constructor(self):
    e = "Constructors and function calls in type annotations are not supported."
    self.check_error("""
      x = ... # type: typing.NamedTuple("A", [])
    """, 1, e)

    self.check_error("""
      x: typing.NamedTuple("A", []) = ...
    """, 1, e)

  def test_match_args(self):
    self.check("""
      from typing import Final

      class A:
          __match_args__ = ("a", "b")

      class B:
          __match_args__: Final = ("a", "b")
    """, """
      class A:
          __match_args__: tuple

      class B:
          __match_args__: tuple
    """)


class QuotedTypeTest(parser_test_base.ParserTestBase):

  def test_annotation(self):
    self.check("""
      class A: ...
      x: "A"
      y: "List[A]" = ...
    """, """
      x: A
      y: List[A] = ...

      class A: ...
    """)

  def test_def(self):
    self.check("""
      def f(x: "int") -> "str": ...
    """, """
      def f(x: int) -> str: ...
    """)


class HomogeneousTypeTest(parser_test_base.ParserTestBase):

  def test_callable_parameters(self):
    self.check("""
      from typing import Callable

      x: Callable[[int, str], bool]""")
    self.check("""
      from typing import Callable

      x = ...  # type: Callable[..., bool]""", """
      from typing import Callable

      x: Callable[..., bool]""")
    self.check("""
      from typing import Any, Callable

      x: Callable[Any, bool]""", """
      from typing import Callable

      x: Callable[..., bool]""")
    self.check("""
      from typing import Any, Callable

      x: Callable[[Any], bool]""")
    self.check("""
      from typing import Callable

      x: Callable[[], bool]""")
    self.check("""
      from typing import Callable

      x = ...  # type: Callable[[int]]""", """
      from typing import Any, Callable

      x: Callable[[int], Any]""")
    self.check("""
      from typing import Callable

      x = ...  # type: Callable[[], ...]""", """
      from typing import Any, Callable

      x: Callable[[], Any]""")
    self.check_error(
        "import typing\n\nx = ...  # type: typing.Callable[int]", 3,
        "First argument to Callable must be a list of argument types")
    self.check_error(
        "import typing\n\nx = ...  # type: typing.Callable[[], bool, bool]", 3,
        "Expected 2 parameters to Callable, got 3")

  def test_ellipsis(self):
    # B[T, ...] becomes B[T].
    self.check("from typing import List\n\nx = ...  # type: List[int, ...]",
               "from typing import List\n\nx: List[int]")
    # Double ellipsis is not allowed.
    self.check_error("x = ...  # type: List[..., ...]", 1,
                     "not supported")
    # Tuple[T] and Tuple[T, ...] are distinct.
    self.check("from typing import Tuple\n\nx = ...  # type: Tuple[int]",
               "from typing import Tuple\n\nx: Tuple[int]")
    self.check("from typing import Tuple\n\nx = ...  # type: Tuple[int, ...]",
               "from typing import Tuple\n\nx: Tuple[int, ...]")

  def test_tuple(self):
    self.check("""
      from typing import Tuple

      x = ...  # type: Tuple[int, str]""",
               """
      from typing import Tuple

      x: Tuple[int, str]""")
    self.check("""
      from typing import Tuple

      x = ...  # type: Tuple[int, str, ...]""",
               """
      from typing import Any, Tuple

      x: Tuple[int, str, Any]""")

  def test_empty_tuple(self):
    self.check("""
      from typing import Tuple

      def f() -> Tuple[()]: ...
    """, """
      from typing import Tuple

      def f() -> Tuple[()]: ...
    """)

  def test_simple(self):
    self.check("x: Foo[int, str]")

  def test_type_tuple(self):
    self.check("x = (str, bytes)",
               "x: tuple")
    self.check("x = (str, bytes,)",
               "x: tuple")
    self.check("x = (str,)",
               "x: tuple")
    self.check("x = str,",
               "x: tuple")


class NamedTupleTest(parser_test_base.ParserTestBase):

  def test_assign_namedtuple(self):
    self.check("X = NamedTuple('X', [])", """
      from typing import NamedTuple

      X = namedtuple_X_0

      class namedtuple_X_0(NamedTuple): ...
    """)

  def test_subclass_namedtuple(self):
    self.check("""
      from typing import NamedTuple
      class X(NamedTuple('X', [])): ...
    """, """
      from typing import NamedTuple

      class namedtuple_X_0(NamedTuple): ...

      class X(namedtuple_X_0): ...
    """)

  def test_trailing_comma(self):
    self.check("""
      from typing import NamedTuple
      Foo = NamedTuple(
          "Foo",
          [
              ("a", int),
              ("b", str),
          ],
      )
    """, """
      from typing import NamedTuple

      Foo = namedtuple_Foo_0

      class namedtuple_Foo_0(NamedTuple):
          a: int
          b: str
    """)

  def test_collections_trailing_comma(self):
    self.check("""
      from collections import namedtuple
      Foo = namedtuple(
        "Foo",
        [
          "a",
          "b",
        ],
      )
    """, """
      from collections import namedtuple
      from typing import Any, NamedTuple

      Foo = namedtuple_Foo_0

      class namedtuple_Foo_0(NamedTuple):
          a: Any
          b: Any
    """)

  def test_collections_namedtuple(self):
    expected = """
      from collections import namedtuple
      from typing import Any, NamedTuple

      X = namedtuple_X_0

      class namedtuple_X_0(NamedTuple):
          y: Any
    """
    self.check("""
      from collections import namedtuple
      X = namedtuple("X", ["y"])
    """, expected)
    self.check("""
      from collections import namedtuple
      X = namedtuple("X", ["y",])
    """, expected)

  def test_typing_namedtuple_class(self):
    self.check("""
      from typing import NamedTuple
      class X(NamedTuple):
        y: int
        z: str
    """, """
      from typing import NamedTuple

      class X(NamedTuple):
          y: int
          z: str
    """)

  def test_typing_namedtuple_class_with_method(self):
    self.check("""
      from typing import NamedTuple
      class X(NamedTuple):
        y: int
        z: str
        def foo(self) -> None: ...
    """, """
      from typing import NamedTuple

      class X(NamedTuple):
          y: int
          z: str
          def foo(self) -> None: ...
    """)

  def test_typing_namedtuple_class_multi_inheritance(self):
    self.check("""
      from typing import NamedTuple
      class X(dict, NamedTuple):
        y: int
        z: str
    """, """
      from typing import NamedTuple

      class X(dict, NamedTuple):
          y: int
          z: str
    """)

  def test_multi_namedtuple_base(self):
    self.check_error("""
      from typing import NamedTuple
      class X(NamedTuple, NamedTuple): ...
    """, 2, "cannot inherit from bare NamedTuple more than once")

  def test_redefine_namedtuple(self):
    self.check("""
      class NamedTuple: ...
    """)


class FunctionTest(parser_test_base.ParserTestBase):

  def test_params(self):
    self.check("def foo() -> int: ...")
    self.check("def foo(x) -> int: ...")
    self.check("def foo(x: int) -> int: ...")
    self.check("def foo(x: int, y: str) -> int: ...")
    # Default values can add type information.
    self.check("def foo(x = 123) -> int: ...",
               "def foo(x: int = ...) -> int: ...")
    self.check("def foo(x = 12.3) -> int: ...",
               "def foo(x: float = ...) -> int: ...")
    self.check("def foo(x = None) -> int: ...",
               "def foo(x = ...) -> int: ...")
    self.check("def foo(x = xyz) -> int: ...",
               "def foo(x = ...) -> int: ...")
    self.check("def foo(x = ...) -> int: ...",
               "def foo(x = ...) -> int: ...")
    # Defaults are ignored if a declared type is present.
    self.check("def foo(x: str = 123) -> int: ...",
               "def foo(x: str = ...) -> int: ...")
    self.check("def foo(x: str = None) -> int: ...",
               "def foo(x: str = ...) -> int: ...")
    # Allow but do not preserve a trailing comma in the param list.
    self.check("def foo(x: int, y: str, z: bool,) -> int: ...",
               "def foo(x: int, y: str, z: bool) -> int: ...")

  def test_star_params(self):
    self.check("def foo(*, x) -> str: ...")
    self.check("def foo(x: int, *args) -> str: ...")
    self.check("def foo(x: int, *args, key: int = ...) -> str: ...")
    self.check("def foo(x: int, *args: float) -> str: ...")
    self.check("def foo(x: int, **kwargs) -> str: ...")
    self.check("def foo(x: int, **kwargs: float) -> str: ...")
    self.check("def foo(x: int, *args, **kwargs) -> str: ...")
    # Various illegal uses of * args.
    self.check_error("def foo(*) -> int: ...", 1,
                     "named arguments must follow bare *")
    self.check_error("def foo(*x, *y) -> int: ...", 1, "invalid syntax")
    self.check_error("def foo(**x, *y) -> int: ...", 1, "invalid syntax")

  @unittest.skip("New parser does not support this syntax")
  def test_ellipsis_param(self):
    self.check("def foo(...) -> int: ...",
               "def foo(*args, **kwargs) -> int: ...")
    self.check("def foo(x: int, ...) -> int: ...",
               "def foo(x: int, *args, **kwargs) -> int: ...")
    self.check_error("def foo(..., x) -> int: ...", 1,
                     "ellipsis (...) must be last parameter")
    self.check_error("def foo(*, ...) -> int: ...", 1,
                     "ellipsis (...) not compatible with bare *")

  def test_typeignore(self):
    self.check("def foo() -> int:  # type: ignore\n  ...",
               "def foo() -> int: ...")
    self.check("def foo() -> int: ...  # type: ignore",
               "def foo() -> int: ...")
    self.check("def foo() -> int: pass  # type: ignore",
               "def foo() -> int: ...")
    self.check("def foo(x) -> int: # type: ignore\n  x=List[int]",
               "def foo(x) -> int:\n    x = List[int]")
    self.check("""
               def foo(x: int,  # type: ignore
                       y: str) -> bool: ...""",
               "def foo(x: int, y: str) -> bool: ...")
    self.check("""
      class Foo:
          bar: str  # type: ignore
    """, """
      class Foo:
          bar: str
    """)
    self.check("""
      class Foo:
          bar = ...  # type: str  # type: ignore
    """, """
      class Foo:
          bar: str
    """)
    self.check("""
      class Foo:
          bar: str = ...  # type: ignore
    """, """
      class Foo:
          bar: str = ...
    """)
    self.check("""
      def f(  # type: ignore
        x: int) -> None: ...
    """, """
      def f(x: int) -> None: ...
    """)

  def test_typeignore_alias(self):
    self.check("""
      class Foo:
          def f(self) -> None: ...
          g = f  # type: ignore
    """, """
      class Foo:
          def f(self) -> None: ...
          def g(self) -> None: ...
    """)

  def test_typeignore_slots(self):
    self.check("""
      class Foo:
          __slots__ = ["a", "b"]  # type: ignore
    """, """
      class Foo:
          __slots__ = ["a", "b"]
    """)

  def test_typeignore_errorcode(self):
    self.check("""
      def f() -> None: ...  # type: ignore[override]
      def g() -> None: ...  # type: ignore[var-annotated]
      def h() -> None: ...  # type: ignore[abstract, no-untyped-def]
    """, """
      def f() -> None: ...
      def g() -> None: ...
      def h() -> None: ...
    """)

  def test_decorators(self):
    # These tests are a bit questionable because most of the decorators only
    # make sense for methods of classes.  But this at least gives us some
    # coverage of the decorator logic.  More sensible tests can be created once
    # classes are implemented.
    self.check("""
      @overload
      def foo() -> int: ...""",
               """
      def foo() -> int: ...""")

    # Accept and disregard type: ignore comments on a decorator
    self.check("""
      @overload
      def foo() -> int: ...
      @overload  # type: ignore  # unsupported signature
      def foo(bool) -> int: ...""",
               """
      from typing import overload

      @overload
      def foo() -> int: ...
      @overload
      def foo(bool) -> int: ...""")

    self.check("""
      @abstractmethod
      def foo() -> int: ...""",
               """
      @abstractmethod
      def foo() -> int: ...""")

    self.check("""
      @abc.abstractmethod
      def foo() -> int: ...""",
               """
      @abstractmethod
      def foo() -> int: ...""")

    self.check("""
      @staticmethod
      def foo() -> int: ...""")

    self.check("""
      @classmethod
      def foo() -> int: ...""")

    self.check("""
      @coroutine
      def foo() -> int: ...""")

    self.check("""
      @asyncio.coroutine
      def foo() -> int: ...""",
               """
      @coroutine
      def foo() -> int: ...""")

    self.check("""
      @asyncio.coroutine
      def foo() -> int: ...
      @coroutines.coroutine
      def foo() -> int: ...
      @coroutine
      def foo() -> str: ...""",
               """
      from typing import overload

      @coroutine
      @overload
      def foo() -> int: ...
      @coroutine
      @overload
      def foo() -> int: ...
      @coroutine
      @overload
      def foo() -> str: ...""")

    self.check_error("""
      def foo() -> str: ...
      @coroutine
      def foo() -> int: ...""",
                     None,
                     "Overloaded signatures for foo disagree on "
                     "coroutine decorators")

    self.check_error("""
      @property
      def foo(self) -> int: ...""",
                     None,
                     "Module-level functions with property decorators: foo")

    self.check_error("""
      @foo.setter
      def foo(self, x) -> int: ...""",
                     None,
                     "Module-level functions with property decorators: foo")

    line = 1 if sys.version_info < (3, 8) else 3
    self.check_error("""
      @classmethod
      @staticmethod
      def foo() -> int: ...""",
                     line,
                     "Too many decorators for foo")

  def test_module_getattr(self):
    self.check("""
      def __getattr__(name) -> int: ...
    """)

    self.check_error("""
      def __getattr__(name) -> int: ...
      def __getattr__(name) -> str: ...
    """, None, "Multiple signatures for module __getattr__")

  def test_type_check_only(self):
    self.check("""
      from typing import type_check_only
      @type_check_only
      def f() -> None: ...
    """, "def f() -> None: ...")

  def test_type_check_only_class(self):
    self.check("""
      from typing import type_check_only
      @type_check_only
      class Foo: ...
    """, "class Foo: ...")

  def test_decorated_class(self):
    self.check("""
      @decorator
      class Foo: ...
    """)

  def test_multiple_class_decorators(self):
    self.check("""
      @decorator1
      @decorator2
      class Foo: ...
    """)

  def test_bad_decorated_class(self):
    line = 1 if sys.version_info < (3, 8) else 2
    self.check_error("""
      @classmethod
      class Foo: ...
    """, line, "Unsupported class decorators: classmethod")

  def test_dataclass_decorator(self):
    self.check("""
      from dataclasses import dataclass

      @dataclass
      class Foo:
          x: int
          y: str = ...
    """)

  def test_dataclass_default_error(self):
    self.check_error("""
      from dataclasses import dataclass
      @dataclass
      class Foo:
        x: int = ...
        y: str
    """, None, "non-default argument y follows default argument x")

  def test_empty_body(self):
    self.check("def foo() -> int: ...")
    self.check("def foo() -> int: ...",
               "def foo() -> int: ...")
    self.check("def foo() -> int: pass",
               "def foo() -> int: ...")
    self.check("""
      def foo() -> int:
        ...""",
               """
      def foo() -> int: ...""")
    self.check("""
      def foo() -> int:
        pass""",
               """
      def foo() -> int: ...""")
    self.check("""
      def foo() -> int:
        '''doc string'''""",
               """
      def foo() -> int: ...""")

  def test_mutators(self):
    # Mutators.
    self.check("""
      def foo(x) -> int:
          x = int""")
    self.check_error("""
      def foo(x) -> int:
          y = int""", 1, "No parameter named 'y'")

  def test_mutator_from_annotation(self):
    self.check("""
      from typing import Generic, TypeVar
      T = TypeVar('T')
      class Foo(Generic[T]):
          def __init__(self: Foo[str]) -> None: ...
    """, """
      from typing import Generic, TypeVar

      T = TypeVar('T')

      class Foo(Generic[T]):
          def __init__(self) -> None:
              self = Foo[str]
    """)

  def test_exceptions(self):
    self.check("""
      def foo(x) -> int:
          raise Error""",
               """
      def foo(x) -> int:
          raise Error()""")
    self.check("""
      def foo(x) -> int:
          raise Error()""")
    self.check("""
      def foo() -> int:
          raise RuntimeError()
          raise TypeError()""")
    self.check("""
      def foo() -> int:
          raise Bar.Error()""", prologue="import Bar")

  def test_invalid_body(self):
    self.check_error("""
      def foo(x) -> int:
        a: str""", 1, "Unexpected statement in function body")

  def test_return(self):
    self.check("def foo() -> int: ...")
    self.check("def foo(): ...",
               "def foo() -> Any: ...",
               prologue="from typing import Any")

  def test_async(self):
    self.check("async def foo() -> int: ...",
               "def foo() -> Coroutine[Any, Any, int]: ...",
               prologue="from typing import Any, Coroutine")


class ClassTest(parser_test_base.ParserTestBase):

  def test_no_bases(self):
    canonical = """
      class Foo: ...
      """

    self.check(canonical, canonical)
    self.check("""
      class Foo():
          pass
      """, canonical)

  def test_bases(self):
    self.check("""
      class Foo(Bar): ...
    """)
    self.check("""
      class Foo(Bar, Baz): ...
      """)

  def test_base_remove_nothingtype(self):
    self.check("""
      class Foo(nothing): ...
      """, """
      class Foo: ...
      """)
    self.check("""
      class Foo(Bar, nothing): ...
      """, """
      class Foo(Bar): ...
      """)

  def test_class_type_ignore(self):
    canonical = """
      class Foo:  # type: ignore
          pass
      class Bar(Foo):  # type: ignore
          pass
      """
    self.check(canonical, """
      class Foo: ...

      class Bar(Foo): ...
    """)

  def test_metaclass(self):
    self.check("""
      class Foo(metaclass=Meta): ...
      """)
    self.check("""
      class Foo(Bar, metaclass=Meta): ...
      """)
    self.check_error("""
      class Foo(badkeyword=Meta): ...
      """, 1, "Unexpected classdef kwarg 'badkeyword'")
    if sys.version_info[:2] >= (3, 10):
      expected_msg = "expected ':'"
    else:
      expected_msg = "positional argument follows keyword argument"
    self.check_error("""
      class Foo(metaclass=Meta, Bar): ...
      """, 1, expected_msg)

  def test_shadow_pep484(self):
    self.check("""
      class List:
          def bar(self) -> List: ...
      """)

  def test_no_body(self):
    canonical = """
      class Foo: ...
      """
    # There are numerous ways to indicate an empty body.
    self.check(canonical, canonical)
    self.check("""
      class Foo(): pass
      """, canonical)
    self.check("""
      class Foo():
          pass
      """, canonical)
    self.check("""
      class Foo():
          ...
      """, canonical)
    # pylint: disable=g-inconsistent-quotes
    self.check('''\
      class Foo():
          """docstring"""
          ...
      ''', canonical)
    self.check('''\
      class Foo():
          """docstring"""
      ''', canonical)
    # Accept type: ignore with empty body
    self.check("""
      class Foo: ...  # type: ignore
      """, canonical)
    self.check("""
      class Foo: # type: ignore
          pass
      """, canonical)

  def test_attribute(self):
    self.check("""
      class Foo:
          a: int
      """)

  def test_method(self):
    self.check("""
      class Foo:
          def a(self, x: int) -> str: ...
      """)

  def test_property(self):
    self.check("""
      class Foo:
          @property
          def a(self) -> int: ...
      """, """
      from typing import Annotated

      class Foo:
          a: Annotated[int, 'property']
      """)

  def test_duplicate_name(self):
    self.check_error("""
      class Foo:
          bar = ...  # type: int
          bar = ...  # type: str
      """, 1, "Duplicate class-level identifier(s): bar")
    self.check_error("""
      class Foo:
          def bar(self) -> int: ...
          bar = ...  # type: str
      """, 1, "Duplicate class-level identifier(s): bar")
    # Multiple method defs are ok (needed for variant signatures).
    self.check("""
      class Foo:
          @overload
          def x(self) -> int: ...
          @overload
          def x(self) -> str: ...
      """, """
      from typing import overload

      class Foo:
          @overload
          def x(self) -> int: ...
          @overload
          def x(self) -> str: ...
      """)

  def test_protocol_base(self):
    self.check("""
      from typing import Protocol

      class Foo(Protocol): ...
    """)

  def test_parameterized_protocol_base(self):
    self.check("""
      from typing import Protocol, TypeVar

      T = TypeVar('T')

      class Foo(Protocol[T]): ...
    """, """
      from typing import Generic, Protocol, TypeVar

      T = TypeVar('T')

      class Foo(Protocol, Generic[T]): ...
    """)

  def test_typing_extensions_parameterized_protocol(self):
    self.check("""
      from typing import TypeVar
      from typing_extensions import Protocol

      T = TypeVar('T')

      class Foo(Protocol[T]): ...
    """, """
      import typing
      from typing import Generic, TypeVar
      from typing_extensions import Protocol

      T = TypeVar('T')

      class Foo(typing.Protocol, Generic[T]): ...
    """)

  def test_bad_typevar_in_mutation(self):
    self.check_error("""
      from typing import Generic, TypeVar

      S = TypeVar('S')
      T = TypeVar('T')
      U = TypeVar('U')
      V = TypeVar('V')

      class Foo(Generic[T]):
        def __init__(self, x: S):
          self = Generic[S, T, U, V]
     """, None, "Type parameter(s) {U, V}")

  def test_nested_class_typing_class_conflict(self):
    ast = parser.parse_string(textwrap.dedent("""
      from typing import Mapping
      class Foo:
        class Mapping: ...
      x: Mapping
    """).lstrip())
    x = ast.Lookup("x")
    self.assertEqual(x.type.name, "typing.Mapping")


class IfTest(parser_test_base.ParserTestBase):

  def test_if_true(self):
    self.check("""
      if sys.version_info >= (3, 5, 0):
        x = ...  # type: int
      """, """
      x: int""")

  def test_if_false(self):
    self.check("""
      if sys.version_info == (1, 2, 3):
        x = ...  # type: int
      """, "")

  def test_else_used(self):
    self.check("""
      if sys.version_info == (1, 2, 3):
        x = ...  # type: int
      else:
        y = ...  # type: str
      """, """
      y: str""")

  def test_else_ignored(self):
    self.check("""
      if sys.version_info >= (3, 5, 0):
        x = ...  # type: int
      else:
        y = ...  # type: str
      """, """
      x: int""")

  def test_elif_used(self):
    self.check("""
      if sys.version_info == (1, 2, 3):
        x = ...  # type: int
      elif sys.version_info >= (3, 5, 0):
        y = ...  # type: float
      else:
        z = ...  # type: str
      """, """
      y: float""")

  def test_elif_preempted(self):
    self.check("""
      if sys.version_info > (1, 2, 3):
        x = ...  # type: int
      elif sys.version_info >= (3, 5, 0):
        y = ...  # type: float
      else:
        z = ...  # type: str
      """, """
      x: int""")

  def test_elif_ignored(self):
    self.check("""
      if sys.version_info == (1, 2, 3):
        x = ...  # type: int
      elif sys.version_info == (4, 5, 6):
        y = ...  # type: float
      else:
        z = ...  # type: str
      """, """
      z: str""")

  def test_nested_if(self):
    self.check("""
      if sys.version_info >= (2, 0):
        if sys.platform == "linux":
          a = ...  # type: int
        else:
          b = ...  # type: int
      else:
        if sys.platform == "linux":
          c = ...  # type: int
        else:
          d = ...  # type: int
      """, "a: int")

  def test_if_or(self):
    self.check("""
      if sys.version_info >= (2, 0) or sys.version_info < (0, 0, 0):
        a = ...  # type: int
      if sys.version_info < (0, 0, 0) or sys.version_info >= (2, 0):
        b = ...  # type: int
      if sys.version_info < (0, 0, 0) or sys.version_info > (4,):
        c = ...  # type: int
      if sys.version_info >= (2, 0) or sys.version_info >= (4, 7):
        d = ...  # type: int
      if (sys.platform == "windows" or sys.version_info < (0,) or
          sys.version_info >= (3, 5)):
        e = ...  # type: int
    """, """
      a: int
      b: int
      d: int
      e: int""")

  def test_if_and(self):
    self.check("""
      if sys.version_info >= (3, 0) and sys.version_info < (4, 0):
        a = ...  # type: int
      if sys.version_info >= (3, 0) and sys.version_info >= (4, 0):
        b = ...  # type: int
    """, """
      a: int""")

  # The remaining tests verify that actions with side effects only take effect
  # within a true block.

  def test_conditional_import(self):
    self.check("""
      if sys.version_info >= (3, 5, 0):
        from foo import Processed
      else:
        from foo import Ignored
      """, "from foo import Processed")

  def test_conditional_alias_or_constant(self):
    self.check("""
      if sys.version_info >= (3, 5, 0):
        x = Processed
      else:
        y = Ignored
      """, "x = Processed")

  def test_conditional_class(self):
    self.check("""
      if sys.version_info >= (3, 5, 0):
        class Processed: ...
      else:
        class Ignored: ...
      """, """
      class Processed: ...
      """)

  def test_conditional_class_registration(self):
    # Class registration allows a local class name to shadow a PEP 484 name.
    # The only time this is noticeable is when the PEP 484 name is one of the
    # capitalized names that gets converted to lower case (i.e. List -> list).
    # In these cases a non-shadowed name would be converted to lower case, and
    # a properly shadowed name would remain capitalized.  In the test below,
    # Dict should be registered, List should not be registered.  Thus after
    # the "if" statement Dict refers to the local Dict class and List refers
    # to the PEP 484 list class.
    self.check("""
      from typing import List
      if sys.version_info >= (3, 5, 0):
        class Dict: ...
      else:
        class List: ...

      x = ...  # type: Dict
      y = ...  # type: List
      """, """
      x: Dict
      y: list

      class Dict: ...
      """)

  def test_conditional_typevar(self):
    # The legacy parser did not handle this correctly - typevars are added
    # regardless of any conditions.
    self.check("""
      if sys.version_info >= (3, 5, 0):
        T = TypeVar('T')
      else:
        F = TypeVar('F')
      """, """
        from typing import TypeVar

        T = TypeVar('T')""")


class ClassIfTest(parser_test_base.ParserTestBase):

  # These tests assume that IfTest has already covered the inner workings of
  # peer's functions.  Instead, they focus on verifying that if statements
  # under a class allow things that normally appear in a class (constants,
  # functions), and disallow statements that aren't allowed in a class (import,
  # etc).

  def test_conditional_constant(self):
    self.check("""
      class Foo:
        if sys.version_info == (3, 4, 0):
          x = ...  # type: int
        elif sys.version_info >= (3, 5, 0):
          y = ...  # type: str
        else:
          z = ...  # type: float
      """, """
      class Foo:
          y: str
      """)

  def test_conditional_method(self):
    self.check("""
      class Foo:
        if sys.version_info == (3, 4, 0):
          def a(self, x: int) -> str: ...
        elif sys.version_info >= (3, 5, 0):
          def b(self, x: int) -> str: ...
        else:
          def c(self, x: int) -> str: ...
      """, """
      class Foo:
          def b(self, x: int) -> str: ...
      """)

  def test_nested(self):
    self.check("""
      class Foo:
        if sys.version_info > (3, 4, 0):
          if sys.version_info >= (3, 5, 0):
            def b(self, x: int) -> str: ...
      """, """
      class Foo:
          def b(self, x: int) -> str: ...
      """)

  def test_no_import(self):
    self.check_error("""
      class Foo:
        if sys.version_info > (3, 4, 0):
          import foo
    """, 3, "Import statements need to be at module level")

  def test_no_class(self):
    self.check("""
      class Foo:
        if sys.version_info <= (3, 4, 0):
          class Bar: ...
    """, """
      class Foo: ...
    """)

  def test_no_typevar(self):
    self.check_error("""
      class Foo:
        if sys.version_info > (3, 4, 0):
          T = TypeVar('T')
    """, 3, r"TypeVars need to be defined at module level")


class ConditionTest(parser_test_base.ParserTestBase):

  def check_cond(self, condition, expected, **kwargs):
    out = "x: int" if expected else ""
    if "version" not in kwargs:
      kwargs["version"] = (3, 6, 5)
    self.check(f"""
      if {condition}:
        x = ...  # type: int
      """, out, **kwargs)

  def check_cond_error(self, condition, message):
    self.check_error(f"""
      if {condition}:
        x = ...  # type: int
      """, 1, message)

  def test_version_eq(self):
    self.check_cond("sys.version_info == (3, 6, 4)", False)
    self.check_cond("sys.version_info == (3, 6, 5)", True)
    self.check_cond("sys.version_info == (3, 6, 6)", False)

  def test_version_ne(self):
    self.check_cond("sys.version_info != (3, 6, 4)", True)
    self.check_cond("sys.version_info != (3, 6, 5)", False)
    self.check_cond("sys.version_info != (3, 6, 6)", True)

  def test_version_lt(self):
    self.check_cond("sys.version_info < (3, 6, 4)", False)
    self.check_cond("sys.version_info < (3, 6, 5)", False)
    self.check_cond("sys.version_info < (3, 6, 6)", True)
    self.check_cond("sys.version_info < (3, 7, 0)", True)

  def test_version_le(self):
    self.check_cond("sys.version_info <= (3, 6, 4)", False)
    self.check_cond("sys.version_info <= (3, 6, 5)", True)
    self.check_cond("sys.version_info <= (3, 6, 6)", True)
    self.check_cond("sys.version_info <= (3, 7, 0)", True)

  def test_version_gt(self):
    self.check_cond("sys.version_info > (3, 6, 0)", True)
    self.check_cond("sys.version_info > (3, 6, 4)", True)
    self.check_cond("sys.version_info > (3, 6, 5)", False)
    self.check_cond("sys.version_info > (3, 6, 6)", False)

  def test_version_ge(self):
    self.check_cond("sys.version_info >= (3, 6, 0)", True)
    self.check_cond("sys.version_info >= (3, 6, 4)", True)
    self.check_cond("sys.version_info >= (3, 6, 5)", True)
    self.check_cond("sys.version_info >= (3, 6, 6)", False)

  def test_version_item(self):
    self.check_cond("sys.version_info[0] == 3", True)

  def test_version_slice(self):
    self.check_cond("sys.version_info[:] == (3, 6, 5)", True)
    self.check_cond("sys.version_info[:2] == (3, 6)", True)
    self.check_cond("sys.version_info[2:] == (5,)", True)
    self.check_cond("sys.version_info[0:1] == (3,)", True)
    self.check_cond("sys.version_info[::] == (3, 6, 5)", True)
    self.check_cond("sys.version_info[1::] == (6, 5)", True)
    self.check_cond("sys.version_info[:2:] == (3, 6)", True)
    self.check_cond("sys.version_info[::-2] == (5, 3)", True)
    self.check_cond("sys.version_info[1:3:] == (6, 5)", True)
    self.check_cond("sys.version_info[1::2] == (6,)", True)
    self.check_cond("sys.version_info[:2:2] == (3,)", True)
    self.check_cond("sys.version_info[3:1:-1] == (5,)", True)

  def test_version_shorter_tuples(self):
    self.check_cond("sys.version_info == (3,)", True, version=(3, 0, 0))
    self.check_cond("sys.version_info == (3, 0)", True, version=(3, 0, 0))
    self.check_cond("sys.version_info == (3, 0, 0)", True, version=(3, 0, 0))
    self.check_cond("sys.version_info == (3,)", False, version=(3, 0, 1))
    self.check_cond("sys.version_info == (3, 0)", False, version=(3, 0, 1))
    self.check_cond("sys.version_info > (3,)", True, version=(3, 0, 1))
    self.check_cond("sys.version_info > (3, 0)", True, version=(3, 0, 1))
    self.check_cond("sys.version_info == (3, 0, 0)", True, version=(3,))
    self.check_cond("sys.version_info == (3, 0, 0)", True, version=(3, 0))

  def test_version_slice_shorter_tuples(self):
    self.check_cond("sys.version_info[:2] == (3,)", True, version=(3, 0, 1))
    self.check_cond("sys.version_info[:2] == (3, 0)", True, version=(3, 0, 1))
    self.check_cond(
        "sys.version_info[:2] == (3, 0, 0)", True, version=(3, 0, 1))
    self.check_cond("sys.version_info[:2] == (3,)", False, version=(3, 1, 0))
    self.check_cond("sys.version_info[:2] == (3, 0)", False, version=(3, 1, 0))
    self.check_cond("sys.version_info[:2] > (3,)", True, version=(3, 1, 0))
    self.check_cond("sys.version_info[:2] > (3, 0)", True, version=(3, 1, 0))
    self.check_cond(
        "sys.version_info[:2] == (3, 0, 0)", True, version=(3,))
    self.check_cond(
        "sys.version_info[:2] == (3, 0, 0)", True, version=(3, 0))

  def test_version_error(self):
    self.check_cond_error('sys.version_info == "foo"',
                          "sys.version_info must be compared to a tuple of "
                          "integers")
    self.check_cond_error("sys.version_info == (1.2, 3)",
                          "sys.version_info must be compared to a tuple of "
                          "integers")
    self.check_cond_error("sys.version_info[0] == 2.0",
                          "an element of sys.version_info must be compared to "
                          "an integer")
    self.check_cond_error("sys.version_info[0] == (2,)",
                          "an element of sys.version_info must be compared to "
                          "an integer")
    self.check_cond_error("sys.version_info[:2] == (2.0, 7)",
                          "sys.version_info must be compared to a tuple of "
                          "integers")
    self.check_cond_error("sys.version_info[:2] == 2",
                          "sys.version_info must be compared to a tuple of "
                          "integers")
    self.check_cond_error("sys.version_info[42] == 42",
                          "tuple index out of range")

  def test_platform_eq(self):
    self.check_cond('sys.platform == "linux"', True)
    self.check_cond('sys.platform == "win32"', False)
    self.check_cond('sys.platform == "foo"', True, platform="foo")

  def test_platform_error(self):
    self.check_cond_error("sys.platform == (1, 2, 3)",
                          "sys.platform must be compared to a string")
    self.check_cond_error('sys.platform < "linux"',
                          "sys.platform must be compared using == or !=")
    self.check_cond_error('sys.platform <= "linux"',
                          "sys.platform must be compared using == or !=")
    self.check_cond_error('sys.platform > "linux"',
                          "sys.platform must be compared using == or !=")
    self.check_cond_error('sys.platform >= "linux"',
                          "sys.platform must be compared using == or !=")

  def test_unsupported_condition(self):
    self.check_cond_error("foo.bar == (1, 2, 3)",
                          "Unsupported condition: 'foo.bar'")


class PropertyDecoratorTest(parser_test_base.ParserTestBase):
  """Tests that cover _parse_signature_as_property()."""

  def test_property_with_type(self):
    expected = """
      from typing import Annotated

      class A:
          name: Annotated[str, 'property']
    """

    # The return type of @property is used for the property type.
    self.check("""
      class A:
          @property
          def name(self) -> str:...
      """, expected)

    self.check("""
      class A:
          @name.setter
          def name(self, value: str) -> None: ...
      """, """
      from typing import Annotated, Any

      class A:
          name: Annotated[Any, 'property']
      """)

    self.check("""
      class A:
          @property
          def name(self) -> str:...

          @name.setter
          def name(self, value: str) -> None: ...
      """, expected)

    self.check("""
      class A:
          @property
          def name(self) -> str:...

          @name.setter
          def name(self, value) -> None: ...
      """, expected)

    self.check("""
      class A:
          @property
          def name(self) -> str:...

          @name.setter
          def name(self, value: int) -> None: ...
      """, expected)

  def test_property_decorator_any_type(self):
    expected = """
          from typing import Annotated, Any

          class A:
              name: Annotated[Any, 'property']
    """

    self.check("""
      class A:
          @property
          def name(self): ...
      """, expected)

    self.check("""
      class A:
          @name.setter
          def name(self, value): ...
      """, expected)

    self.check("""
      class A:
          @name.deleter
          def name(self): ...
      """, expected)

    self.check("""
      class A:
          @name.setter
          def name(self, value): ...

          @name.deleter
          def name(self): ...
      """, expected)

  def test_property_decorator_bad_syntax(self):
    line = 1 if sys.version_info < (3, 8) else 2
    self.check_error("""
      class A:
          @property
          def name(self, bad_arg): ...
      """, 1, "@property needs 1 param(s), got 2")

    self.check_error("""
      class A:
          @name.setter
          def name(self): ...
      """, 1, "@name.setter needs 2 param(s), got 1")

    line = 2 if sys.version_info < (3, 8) else 4
    self.check_error("""
      class A:
          @property
          @staticmethod
          def name(self): ...
      """, line, "Too many decorators for name")

    self.check_error("""
      @property
      def name(self): ...
      """, None, "Module-level functions with property decorators: name")

  def test_property_clash(self):
    self.check_error("""
      class A:
          @property
          def name(self) -> str: ...

          @property
          def name(self) -> int: ...
      """, 1, "Invalid property decorators for method `name`")

  def test_abstract_property(self):
    self.check("""
      class Foo:
        @property
        @abstractmethod
        def x(self) -> int: ...
        @x.setter
        def x(self, y: int) -> None: ...
    """, """
      from typing import Annotated

      class Foo:
          x: Annotated[int, 'property']
    """)


class MergeSignaturesTest(parser_test_base.ParserTestBase):

  def test_property(self):
    self.check("""
      class A:
          @property
          def name(self) -> str: ...
      """, """
      from typing import Annotated

      class A:
          name: Annotated[str, 'property']
      """)

  def test_method(self):
    self.check("""
      class A:
          def name(self) -> str: ...
      """)

  def test_merged_method(self):
    ast = self.check("""
      def foo(x: int) -> str: ...
      def foo(x: str) -> str: ...""",
                     """
      from typing import overload

      @overload
      def foo(x: int) -> str: ...
      @overload
      def foo(x: str) -> str: ...""")
    self.assertEqual(len(ast.functions), 1)
    foo = ast.functions[0]
    self.assertEqual(len(foo.signatures), 2)

  def test_method_and_property_error(self):
    self.check_error("""
      class A:
          @property
          def name(self): ...

          def name(self): ...
      """, 1, "Overloaded signatures for name disagree on decorators")

  def test_overloaded_signatures_disagree(self):
    self.check_error("""
      class A:
          @staticmethod
          def foo(x: int): ...
          @classmethod
          def foo(x: str): ...
      """, 1, "Overloaded signatures for foo disagree on decorators")

  def test_classmethod(self):
    ast = self.check("""
      class A:
          @classmethod
          def foo(x: int) -> str: ...
      """)
    self.assertEqual(pytd.MethodKind.CLASSMETHOD,
                     ast.classes[0].methods[0].kind)

  def test_staticmethod(self):
    ast = self.check("""
      class A:
          @staticmethod
          def foo(x: int) -> str: ...
      """)
    self.assertEqual(pytd.MethodKind.STATICMETHOD,
                     ast.classes[0].methods[0].kind)

  def test_new(self):
    ast = self.check("""
      class A:
          def __new__(self) -> A: ...
      """)
    self.assertEqual(pytd.MethodKind.STATICMETHOD,
                     ast.classes[0].methods[0].kind)

  def test_abstractmethod(self):
    ast = self.check("""
      class A:
          @abstractmethod
          def foo(x: int) -> str: ...
      """)
    self.assertEqual(pytd.MethodKind.METHOD, ast.Lookup("A").Lookup("foo").kind)
    self.assertEqual(True, ast.Lookup("A").Lookup("foo").is_abstract)

  def test_abstractmethod_manysignatures(self):
    ast = self.check("""
      class A:
          @abstractmethod
          def foo(x: int) -> str: ...
          @abstractmethod
          def foo(x: int, y: int) -> str: ...
          @abstractmethod
          def foo(x: int, y: int, z: int) -> str: ...
      """, """
      from typing import overload

      class A:
          @abstractmethod
          @overload
          def foo(x: int) -> str: ...
          @abstractmethod
          @overload
          def foo(x: int, y: int) -> str: ...
          @abstractmethod
          @overload
          def foo(x: int, y: int, z: int) -> str: ...
      """)
    self.assertEqual(pytd.MethodKind.METHOD, ast.Lookup("A").Lookup("foo").kind)
    self.assertEqual(True, ast.Lookup("A").Lookup("foo").is_abstract)

  def test_abstractmethod_conflict(self):
    self.check_error("""
      class A:
          @abstractmethod
          def foo(x: int) -> str: ...
          def foo(x: int, y: int) -> str: ...
      """, 1, "Overloaded signatures for foo disagree on "
                     "abstractmethod decorators")


class AnyTest(parser_test_base.ParserTestBase):

  def test_generic_any(self):
    self.check("""
      from typing import Any
      x = ...  # type: Any[int]""",
               """
      from typing import Any

      x: Any""")

  def test_generic_any_alias(self):
    self.check("""
      from typing import Any
      Foo = Any
      Bar = Foo[int]
      x = ...  # type: Bar[int, str]""",
               """
      from typing import Any

      Foo = Any
      Bar = Any

      x: Any""")


class CanonicalPyiTest(parser_test_base.ParserTestBase):

  def test_canonical_version(self):
    src = textwrap.dedent("""
        from typing import Any
        def foo(x: int = 0) -> Any: ...
        def foo(x: str) -> Any: ...
    """)
    expected = textwrap.dedent("""
        from typing import Any, overload

        @overload
        def foo(x: int = ...) -> Any: ...
        @overload
        def foo(x: str) -> Any: ...
    """).strip()
    self.assertMultiLineEqual(
        parser.canonical_pyi(src, options=self.options), expected)


class TypeMacroTest(parser_test_base.ParserTestBase):

  def test_simple(self):
    self.check("""
      from typing import List, TypeVar
      Alias = List[List[T]]
      T = TypeVar('T')
      S = TypeVar('S')
      def f(x: Alias[S]) -> S: ...
      def g(x: Alias[str]) -> str: ...""", """
      from typing import List, TypeVar

      Alias = List[List[T]]

      S = TypeVar('S')
      T = TypeVar('T')

      def f(x: List[List[S]]) -> S: ...
      def g(x: List[List[str]]) -> str: ...""")

  def test_partial_replacement(self):
    self.check("""
      from typing import Dict, TypeVar
      DictAlias = Dict[int, V]
      V = TypeVar('V')
      def f(x: DictAlias[str]) -> None: ...""", """
      from typing import Dict, TypeVar

      DictAlias = Dict[int, V]

      V = TypeVar('V')

      def f(x: Dict[int, str]) -> None: ...""")

  def test_multiple_parameters(self):
    self.check("""
      from typing import Dict, List, TypeVar
      Alias = List[Dict[K, V]]
      K = TypeVar('K')
      V = TypeVar('V')
      def f(x: Alias[K, V]) -> Dict[K, V]: ...""", """
      from typing import Dict, List, TypeVar

      Alias = List[Dict[K, V]]

      K = TypeVar('K')
      V = TypeVar('V')

      def f(x: List[Dict[K, V]]) -> Dict[K, V]: ...""")

  def test_union(self):
    self.check("""
      from typing import List, TypeVar, Union
      Alias = Union[List[T], List[S]]
      T = TypeVar('T')
      S = TypeVar('S')
      def f(x: Alias[S, T]) -> Union[S, T]: ...""", """
      from typing import List, TypeVar, Union

      Alias = Union[List[T], List[S]]

      S = TypeVar('S')
      T = TypeVar('T')

      def f(x: Union[List[S], List[T]]) -> Union[S, T]: ...""")

  def test_repeated_type_parameter(self):
    self.check("""
      from typing import Dict, TypeVar
      Alias = Dict[T, T]
      T = TypeVar('T')
      def f(x: Alias[str]) -> None: ...""", """
      from typing import Dict, TypeVar

      Alias = Dict[T, T]

      T = TypeVar('T')

      def f(x: Dict[str, str]) -> None: ...""")

  def test_wrong_parameter_count(self):
    self.check_error("""
      from typing import List, TypeVar
      Alias = List[List[T]]
      T = TypeVar('T')
      def f(x: Alias[T, T]) -> T: ...
    """, 4, "List[List[T]] expected 1 parameters, got 2")

  def test_anystr(self):
    self.check("""
      from typing import AnyStr, List
      Alias = List[AnyStr]
      def f(x: Alias[str]) -> None: ...
    """, """
      from typing import AnyStr, List

      Alias = List[AnyStr]

      def f(x: List[str]) -> None: ...
    """)


class ImportTypeIgnoreTest(parser_test_base.ParserTestBase):

  def test_import(self):
    self.check("""
      import mod  # type: ignore
      def f(x: mod.attr) -> None: ...
    """, """
      import mod

      def f(x: mod.attr) -> None: ...""")

  def test_from_import(self):
    src = textwrap.dedent("""
      from mod import attr  # type: ignore
      def f(x: attr) -> None: ...
    """)
    ast = parser.parse_string(src, options=self.options)
    self.assertTrue(ast.Lookup("attr"))
    self.assertTrue(ast.Lookup("f"))

  def test_relative_import(self):
    src = textwrap.dedent("""
      from . import attr  # type: ignore
      def f(x: attr) -> None: ...
    """)
    ast = parser.parse_string(src, options=self.options)
    self.assertTrue(ast.Lookup("attr"))
    self.assertTrue(ast.Lookup("f"))

  def test_relative_import_parent(self):
    src = textwrap.dedent("""
      from .. import attr  # type: ignore
      def f(x: attr) -> None: ...
    """)
    ast = parser.parse_string(src, options=self.options)
    self.assertTrue(ast.Lookup("attr"))
    self.assertTrue(ast.Lookup("f"))


class LiteralTest(parser_test_base.ParserTestBase):

  def test_bool(self):
    self.check("""
      from typing import Literal

      x: Literal[False]
      y: Literal[True]
    """)

  def test_int(self):
    self.check("""
      from typing import Literal

      x: Literal[42]
    """)

  def test_string(self):
    self.check("""
      from typing import Literal

      x: Literal['x']
      y: Literal['']
    """)

  def test_bytestring(self):
    self.check("""
      from typing import Literal

      x: Literal[b'']
      y: Literal[b'']
      z: Literal[b'xyz']
    """)

  @test_utils.skipFromPy((3, 8), "No longer tagging strings 'unicode'")
  def test_unicodestring(self):
    self.check("""
      from typing import Literal

      y: Literal[u'']
      z: Literal[u'xyz']
    """)

  def test_none(self):
    self.check("""
      from typing import Literal

      x: Literal[None]
    """, "x: None")

  def test_enum(self):
    self.check("""
      import enum
      from typing import Literal

      x: Literal[Color.RED]

      class Color(enum.Enum):
          RED: str
    """)

  def test_multiple_parameters(self):
    self.check("""
      from typing import Literal

      x: Literal[True, 0, b"", "", None]
    """, """
      from typing import Literal, Optional

      x: Optional[Literal[True, 0, b'', '']]
    """)

  def test_stray_number(self):
    self.check_error("""
      from typing import Tuple

      x: Tuple[int, int, 0, int]
    """, 3, "Tuple[_, _, 0, _] not supported")

  def test_stray_string(self):
    self.check_error("""
      from typing import Tuple

      x: Tuple[str, str, '', str]
    """, 3, "Tuple[_, _, '', _] not supported")

  def test_stray_bytestring(self):
    self.check_error("""
      from typing import Tuple

      x: Tuple[str, b'', str, str]
    """, 3, "Tuple[_, b'', _, _] not supported")

  @test_utils.skipFromPy((3, 8), "No longer tagging strings 'unicode'")
  def test_stray_unicodestring(self):
    self.check_error("""
      from typing import Tuple

      x: Tuple[str, u'', str, str]
    """, 3, "Tuple[_, u'', _, _] not supported")

  def test_typing_extensions(self):
    self.check("""
      from typing_extensions import Literal

      x: Literal[42]
    """)

  def test_unnest(self):
    self.check("""
      from typing import Literal
      MyLiteralAlias = Literal[42]
      x: Literal[MyLiteralAlias, Literal[Literal[True]], None]
      y: Literal[1, Literal[2, Literal[3]]]
    """, """
      from typing import Literal, Optional

      MyLiteralAlias = Literal[42]

      x: Optional[Literal[42, True]]
      y: Literal[1, 2, 3]
    """)

  def test_bad_value(self):
    self.check_error("""
      from typing import Literal
      x: Literal[0.0]
    """, 2, "Invalid type `float` in Literal[0.0].")

  def test_forbid_expressions(self):
    msg = "Expressions are not allowed in typing.Literal."
    self.check_error("""
      from typing import Literal
      x: Literal[3+4]
    """, 2, msg)
    self.check_error("""
      from typing_extensions import Literal
      x: Literal[3+4]
    """, 2, msg)
    self.check_error("""
      import typing
      x: typing.Literal[3+4]
    """, 2, msg)
    self.check_error("""
      import typing_extensions
      x: typing_extensions.Literal[3+4]
    """, 2, msg)

  def test_final_literals(self):
    # See https://github.com/python/typeshed/issues/7258 for context.
    self.check("""
      import enum
      from typing_extensions import Final

      class Color(enum.Enum):
        RED: str

      x1: Final = 3
      x2: Final = True
      x3: Final = 'x3'
      x4: Final = b'x4'
      x5: Final = None
      x6: Final = Color.RED
    """, """
      import enum
      from typing import Literal
      from typing_extensions import Final

      x1: Literal[3]
      x2: Literal[True]
      x3: Literal['x3']
      x4: Literal[b'x4']
      x5: Literal[None]
      x6: Literal[Color.RED]

      class Color(enum.Enum):
          RED: str
    """)

  def test_final_non_literal(self):
    self.check("""
      from typing_extensions import Final
      x: Final
      y: Final = ...
    """, """
      import typing_extensions
      from typing_extensions import Final

      x: typing_extensions.Final
      y: typing_extensions.Final = ...
    """)

  def test_bad_final_literal(self):
    msg = ("Default value for x: Final can only be '...' or a legal Literal "
           "parameter, got LITERAL(3.14)")
    self.check_error("""
      from typing_extensions import Final
      x: Final = 3.14
    """, 2, msg)


class TypedDictTest(parser_test_base.ParserTestBase):

  def test_assign(self):
    self.check("""
      from typing_extensions import TypedDict
      X = TypedDict('X', {})
    """, """
      from typing import Any, Dict
      from typing_extensions import TypedDict

      X = Dict[str, Any]
    """)

  def test_assign_with_items(self):
    self.check("""
      from typing_extensions import TypedDict
      X = TypedDict('X', {'a': int, 'b': str})
    """, """
      from typing import Any, Dict
      from typing_extensions import TypedDict

      X = Dict[str, Any]
    """)

  def test_assign_with_kwarg(self):
    self.check("""
      from typing_extensions import TypedDict
      X = TypedDict('X', {}, total=False)
    """, """
      from typing import Any, Dict
      from typing_extensions import TypedDict

      X = Dict[str, Any]
    """)

  def test_trailing_comma(self):
    self.check("""
      from typing_extensions import TypedDict
      X = TypedDict('X', {
          'a': int,
          'b': str,
      },)
    """, """
      from typing import Any, Dict
      from typing_extensions import TypedDict

      X = Dict[str, Any]
    """)

  def test_kwarg(self):
    self.check("""
      from typing import TypedDict

      class Foo(TypedDict, total=False): ...
    """, """
      from typing import TypedDict

      class Foo(TypedDict): ...
    """)

  def test_typing_extensions(self):
    self.check("""
      from typing_extensions import TypedDict

      class Foo(TypedDict, total=False): ...
    """, """
      import typing_extensions
      from typing_extensions import TypedDict

      class Foo(typing_extensions.TypedDict): ...
    """)

  def test_multiple_classdef_kwargs(self):
    self.check("""
      from typing import TypedDict

      class Foo(TypedDict, total=False, metaclass=Meta): ...
    """, """
      from typing import TypedDict

      class Foo(TypedDict, metaclass=Meta): ...
    """)

  def test_total_in_subclass(self):
    self.check("""
      from typing import TypedDict
      class Foo(TypedDict):
          x: str
      class Bar(Foo, total=False):
          y: int
    """, """
      from typing import TypedDict

      class Foo(TypedDict):
          x: str

      class Bar(Foo):
          y: int
    """)


class NewTypeTest(parser_test_base.ParserTestBase):

  def test_basic(self):
    self.check("""
      from typing import NewType
      X = NewType('X', int)
    """, """
      X = newtype_X_0

      class newtype_X_0(int):
          def __init__(self, val: int) -> None: ...
    """)

  def test_fullname(self):
    self.check("""
      import typing
      X = typing.NewType('X', int)
    """, """
      import typing

      X = newtype_X_0

      class newtype_X_0(int):
          def __init__(self, val: int) -> None: ...
    """)


class MethodAliasTest(parser_test_base.ParserTestBase):

  def test_normal_method(self):
    self.check("""
      class Foo:
          def f(self, x: int) -> None: ...
      _foo: Foo
      f1 = Foo.f
      f2 = _foo.f
    """, """
      _foo: Foo

      class Foo:
          def f(self, x: int) -> None: ...

      def f1(self, x: int) -> None: ...
      def f2(x: int) -> None: ...
    """)

  def test_classmethod(self):
    self.check("""
      class Foo:
          @classmethod
          def f(cls, x: int) -> None: ...
      _foo: Foo
      f1 = Foo.f
      f2 = _foo.f
    """, """
      _foo: Foo

      class Foo:
          @classmethod
          def f(cls, x: int) -> None: ...

      def f1(x: int) -> None: ...
      def f2(x: int) -> None: ...
    """)

  def test_staticmethod(self):
    self.check("""
      class Foo:
          @staticmethod
          def f(x: int) -> None: ...
      _foo: Foo
      f1 = Foo.f
      f2 = _foo.f
    """, """
      _foo: Foo

      class Foo:
          @staticmethod
          def f(x: int) -> None: ...

      def f1(x: int) -> None: ...
      def f2(x: int) -> None: ...
    """)

  def test_nested_constant(self):
    self.check("""
      class Foo:
          foo: Foo
          def f(self, x: int) -> None: ...
      f = Foo.foo.f
    """, """
      class Foo:
          foo: Foo
          def f(self, x: int) -> None: ...

      def f(x: int) -> None: ...
    """)


class AnnotatedTest(parser_test_base.ParserTestBase):
  """Test typing.Annotated."""

  def test_annotated(self):
    self.check("""
      from typing import Annotated

      class Foo:
          x: Annotated[int, 'a', 'b', 'c']
    """)

  def test_annotated_from_extensions(self):
    self.check("""
      from typing_extensions import Annotated

      class Foo:
          x: Annotated[int, 'a', 'b', 'c']
    """)

  def test_dict(self):
    self.check("""
      from typing_extensions import Annotated

      class Foo:
          x: Annotated[int, {'a': 'A', 'b': True, 'c': Foo}]
    """, """
      from typing_extensions import Annotated

      class Foo:
          x: Annotated[int, {'a': 'A', 'b': True, 'c': 'Foo'}]
    """)

  def test_call(self):
    self.check("""
      from typing_extensions import Annotated

      class Foo:
          x: Annotated[int, Deprecated("use new api")]
          y: Annotated[int, unit('s', exp=9)]
    """, """
      from typing_extensions import Annotated

      class Foo:
          x: Annotated[int, {'tag': 'Deprecated', 'reason': 'use new api'}]
          y: Annotated[int, {'tag': 'call', 'fn': 'unit', 'posargs': ('s',), 'kwargs': {'exp': 9}}]
    """)


class ErrorTest(test_base.UnitTest):
  """Test parser errors."""

  def test_filename(self):
    src = textwrap.dedent("""
      a: int
      a: int
    """)
    with self.assertRaisesRegex(parser.ParseError, "File.*foo.pyi"):
      parser.parse_pyi(src, "foo.pyi", "foo")

  def test_lineno(self):
    src = textwrap.dedent("""
      class A:
        __slots__ = 0
    """)
    with self.assertRaisesRegex(parser.ParseError, "line 3"):
      parser.parse_pyi(src, "foo.py", "foo")


class ParamsTest(test_base.UnitTest):
  """Test input parameter handling."""

  def test_feature_version(self):
    cases = [
        [(3,), sys.version_info.minor],
        [(3, 7), 7],
        [(3, 8, 2), 8]
    ]
    for version, expected in cases:
      actual = parser._feature_version(version)
      self.assertEqual(actual, expected)


class ParamSpecTest(parser_test_base.ParserTestBase):

  def test_from_typing(self):
    self.check("""
      from typing import Awaitable, Callable, ParamSpec, TypeVar

      P = ParamSpec('P')
      R = TypeVar('R')

      def f(x: Callable[P, R]) -> Callable[P, Awaitable[R]]: ...
    """, """
      from typing import Awaitable, Callable, TypeVar

      R = TypeVar('R')

      def f(x: Callable[..., R]) -> Callable[..., Awaitable[R]]: ...
    """)

  def test_from_typing_extensions(self):
    self.check("""
      from typing import Awaitable, Callable, TypeVar
      from typing_extensions import ParamSpec

      P = ParamSpec('P')
      R = TypeVar('R')

      def f(x: Callable[P, R]) -> Callable[P, Awaitable[R]]: ...
    """, """
      from typing import Awaitable, Callable, TypeVar
      from typing_extensions import ParamSpec

      R = TypeVar('R')

      def f(x: Callable[..., R]) -> Callable[..., Awaitable[R]]: ...
    """)

  def test_custom_generic(self):
    self.check("""
      from typing import Callable, Generic, ParamSpec, TypeVar

      P = ParamSpec('P')
      T = TypeVar('T')

      class X(Generic[T, P]):
          f: Callable[P, int]
          x: T
    """, """
      from typing import Callable, Generic, TypeVar

      P = TypeVar('P')
      T = TypeVar('T')

      class X(Generic[T, P]):
          f: Callable[..., int]
          x: T
    """)

  def test_use_custom_generic(self):
    self.check("""
      from typing import Callable, Generic, TypeVar
      from typing_extensions import ParamSpec

      _T = TypeVar('_T')
      _P = ParamSpec('_P')

      class Foo(Generic[_P, _T]): ...

      def f(x: Callable[_P, _T]) -> Foo[_P, _T]: ...
    """, """
      from typing import Any, Callable, Generic, TypeVar
      from typing_extensions import ParamSpec

      _P = TypeVar('_P')
      _T = TypeVar('_T')

      class Foo(Generic[_P, _T]): ...

      def f(x: Callable[..., _T]) -> Foo[Any, _T]: ...
    """)

  @test_base.skip("ParamSpec in custom generic classes not supported yet")
  def test_double_brackets(self):
    # Double brackets can be omitted when instantiating a class parameterized
    # with only a single ParamSpec.
    self.check("""
      from typing import Generic, ParamSpec

      P = ParamSpec('P')

      class X(Generic[P]): ...

      def f1(x: X[int, str]) -> None: ...
      def f2(x: X[[int, str]]) -> None: ...
    """, """
      from typing import Generic, ParamSpec

      P = ParamSpec('P')

      class X(Generic[P]): ...

      def f1(x: X[int, str]) -> None: ...
      def f2(x: X[int, str]) -> None: ...
    """)

  def test_paramspec_args(self):
    self.check("""
      from typing import Callable, ParamSpec, TypeVar
      P = ParamSpec('P')
      T = TypeVar('T')
      def f(x: Callable[P, T], *args: P.args, **kwargs: P.kwargs) -> T: ...
    """, """
      from typing import Callable, TypeVar

      T = TypeVar('T')

      def f(x: Callable[..., T], *args, **kwargs) -> T: ...
    """)

  def test_paramspec_args_error(self):
    self.check_error("""
      from typing import Any, Callable
      from typing_extensions import ParamSpec
      _P = ParamSpec("_P")

      class Foo:
        def __init__(
            self, func: Callable[_P, Any], args: _P.args, kwds: _P.kwds
        ) -> None: ...
    """, 7, "Unrecognized ParamSpec attribute: kwds")

  def test_two_classes(self):
    self.check("""
      from typing import Generic, ParamSpec
      P = ParamSpec('P')
      class C1(Generic[P]): ...
      class C2(Generic[P]): ...
    """, """
      from typing import Generic, TypeVar

      P = TypeVar('P')

      class C1(Generic[P]): ...

      class C2(Generic[P]): ...
    """)


class ConcatenateTest(parser_test_base.ParserTestBase):

  def test_from_typing(self):
    self.check("""
      from typing import Callable, Concatenate, ParamSpec, TypeVar

      P = ParamSpec('P')
      R = TypeVar('R')

      class X: ...

      def f(x: Callable[Concatenate[X, P], R]) -> Callable[P, R]: ...
    """, """
      from typing import Callable, TypeVar

      R = TypeVar('R')

      class X: ...

      def f(x: Callable[..., R]) -> Callable[..., R]: ...
    """)

  def test_from_typing_extensions(self):
    self.check("""
      from typing import Callable, TypeVar
      from typing_extensions import Concatenate, ParamSpec

      P = ParamSpec('P')
      R = TypeVar('R')

      class X: ...

      def f(x: Callable[Concatenate[X, P], R]) -> Callable[P, R]: ...
    """, """
      from typing import Callable, TypeVar
      from typing_extensions import Concatenate
      from typing_extensions import ParamSpec

      R = TypeVar('R')

      class X: ...

      def f(x: Callable[..., R]) -> Callable[..., R]: ...
    """)


class UnionOrTest(parser_test_base.ParserTestBase):

  def test_basic(self):
    self.check("""
      def f(x: int | str) -> None: ...
      def g(x: bool | str | float) -> None: ...
      def h(x: str | None) -> None: ...
    """, """
      from typing import Optional, Union

      def f(x: Union[int, str]) -> None: ...
      def g(x: Union[bool, str, float]) -> None: ...
      def h(x: Optional[str]) -> None: ...
    """)


class TypeGuardTest(parser_test_base.ParserTestBase):

  def test_typing_extensions(self):
    self.check("""
      from typing import List
      from typing_extensions import TypeGuard

      def f(x: List[object]) -> TypeGuard[List[str]]: ...
    """, """
      from typing import List
      from typing_extensions import TypeGuard

      def f(x: List[object]) -> bool: ...
  """)

  def test_typing(self):
    self.check("""
      from typing import List, TypeGuard

      def f(x: List[object]) -> TypeGuard[List[str]]: ...
    """, """
      from typing import List

      def f(x: List[object]) -> bool: ...
    """)


class AllTest(parser_test_base.ParserTestBase):

  def check(self, src, expected):
    tree = self.parse(src)
    all_ = [x for x in tree.constants if x.name == "__all__"]
    pyval = all_[0].value if all_ else None
    self.assertEqual(pyval, expected)

  def test_basic(self):
    self.check("""
      __all__ = ["f", "g"]
    """, ("f", "g"))

  def test_tuple(self):
    self.check("""
    __all__ = ("f", "g")
    """, ("f", "g"))

  def test_augment(self):
    self.check("""
      __all__ = ["f", "g"]
      __all__ += ["h"]
    """, ("f", "g", "h"))

  def test_if(self):
    self.check("""
      __all__ = ["f", "g"]
      if sys.version_info > (3, 6, 0):
        __all__ += ["h"]
    """, ("f", "g", "h"))

  def test_else(self):
    self.check("""
      __all__ = ["f", "g"]
      if sys.version_info < (3, 6, 0):
        __all__ += ["e"]
      else:
        __all__ += ["h"]
    """, ("f", "g", "h"))


if __name__ == "__main__":
  unittest.main()



"""Debug utils for working with the indexer and the AST."""

from pytype.ast import debug

# pylint: disable=protected-access
# We never care about protected access when writing debug code!


def format_loc(location):
  # location is (line, column)
  fmt = "%d:%2d" % location
  return fmt.rjust(8)


def format_def_with_location(defn, loc):
  return f"{format_loc(loc)}  | {defn.typ.ljust(15)} {defn.format()}"


def format_ref(ref):
  return (f"{format_loc(ref.location)}  | {ref.typ.ljust(15)}  "
          f"{ref.scope}.{ref.name}")


def format_call(call):
  return f"{format_loc(call.location)}  | {'Call'.ljust(15)}  {call.func}"


def typename(node):
  return node.__class__.__name__


def show_defs(index):
  """Show definitions."""
  for def_id in index.locs:
    defn = index.defs[def_id]
    for loc in index.locs[def_id]:
      print(format_def_with_location(defn, loc.location))
      if defn.doc:
        print(" "*28 + str(defn.doc))


def show_refs(index):
  """Show references and associated definitions."""
  indent = "          :  "
  for ref, defn in index.links:
    print(format_ref(ref))
    if defn:
      print(indent, defn.format())
      for loc in index.locs[defn.id]:
        print(indent, format_def_with_location(defn, loc.location))
    else:
      print(indent, "None")
    continue


def show_calls(index):
  for call in index.calls:
    print(format_call(call))


def display_type(data):
  """Convert a pytype internal type to a display type."""
  name = "typing.Any"
  if data and data[0]:
    d = data[0][0]
    if d.cls:
      name = d.cls.full_name
  if name == "unsolveable":
    name = "typing.Any"
  elif name.startswith("z__pytype_partial"):
    name = "<unknown>"
  return name


def show_types(index):
  """Show inferred types."""
  out = []
  for def_id in index.locs:
    defn = index.defs[def_id]
    for loc in index.locs[def_id]:
      out.append((loc.location, defn.typ, defn.name, display_type(defn.data)))
  for ref, defn in index.links:
    out.append((ref.location, defn.typ, ref.name, display_type(ref.data)))
  # Sort by location
  for location, category, name, typ in sorted(out, key=lambda x: x[0]):
    # Filter out some noise
    if (category in ("FunctionDef", "IsInstance") or
        typ in ("builtins.module", "__future__._Feature")):
      continue
    print(f"{format_loc(location)}  |  {name.ljust(35)}  {typ}")


def show_index(index):
  """Display output in human-readable format."""

  def separator():
    print("\n--------------------\n")

  show_defs(index)
  separator()
  show_refs(index)
  separator()
  show_calls(index)
  separator()


def show_map(name, mapping):
  print("%s: {" % name)
  for k, v in mapping.items():
    print("  ", k, v)
  print("}")


# reexport AST dumper

dump = debug.dump



from _typeshed import Self
from typing import Any, Callable, Generic, Iterable, Iterator, Sequence, TypeVar, overload

_T = TypeVar("_T")
_S = TypeVar("_S")

def count(start: int = ..., step: int = ...) -> Iterator[int]: ...  # more general types?

class cycle(Iterator[_T], Generic[_T]):
    def __init__(self, iterable: Iterable[_T]) -> None: ...
    def next(self) -> _T: ...
    def __iter__(self: Self) -> Self: ...

def repeat(object: _T, times: int = ...) -> Iterator[_T]: ...

class chain(Iterator[_T], Generic[_T]):
    def __init__(self, *iterables: Iterable[_T]) -> None: ...
    def next(self) -> _T: ...
    def __iter__(self: Self) -> Self: ...
    @staticmethod
    def from_iterable(iterable: Iterable[Iterable[_S]]) -> Iterator[_S]: ...

def compress(data: Iterable[_T], selectors: Iterable[Any]) -> Iterator[_T]: ...
def dropwhile(predicate: Callable[[_T], Any], iterable: Iterable[_T]) -> Iterator[_T]: ...
def ifilter(predicate: Callable[[_T], Any] | None, iterable: Iterable[_T]) -> Iterator[_T]: ...
def ifilterfalse(predicate: Callable[[_T], Any] | None, iterable: Iterable[_T]) -> Iterator[_T]: ...
@overload
def groupby(iterable: Iterable[_T], key: None = ...) -> Iterator[tuple[_T, Iterator[_T]]]: ...
@overload
def groupby(iterable: Iterable[_T], key: Callable[[_T], _S]) -> Iterator[tuple[_S, Iterator[_T]]]: ...
@overload
def islice(iterable: Iterable[_T], stop: int | None) -> Iterator[_T]: ...
@overload
def islice(iterable: Iterable[_T], start: int | None, stop: int | None, step: int | None = ...) -> Iterator[_T]: ...

_T1 = TypeVar("_T1")
_T2 = TypeVar("_T2")
_T3 = TypeVar("_T3")
_T4 = TypeVar("_T4")
_T5 = TypeVar("_T5")
_T6 = TypeVar("_T6")

@overload
def imap(func: Callable[[_T1], _S], iter1: Iterable[_T1]) -> Iterator[_S]: ...
@overload
def imap(func: Callable[[_T1, _T2], _S], iter1: Iterable[_T1], iter2: Iterable[_T2]) -> Iterator[_S]: ...
@overload
def imap(
    func: Callable[[_T1, _T2, _T3], _S], iter1: Iterable[_T1], iter2: Iterable[_T2], iter3: Iterable[_T3]
) -> Iterator[_S]: ...
@overload
def imap(
    func: Callable[[_T1, _T2, _T3, _T4], _S],
    iter1: Iterable[_T1],
    iter2: Iterable[_T2],
    iter3: Iterable[_T3],
    iter4: Iterable[_T4],
) -> Iterator[_S]: ...
@overload
def imap(
    func: Callable[[_T1, _T2, _T3, _T4, _T5], _S],
    iter1: Iterable[_T1],
    iter2: Iterable[_T2],
    iter3: Iterable[_T3],
    iter4: Iterable[_T4],
    iter5: Iterable[_T5],
) -> Iterator[_S]: ...
@overload
def imap(
    func: Callable[[_T1, _T2, _T3, _T4, _T5, _T6], _S],
    iter1: Iterable[_T1],
    iter2: Iterable[_T2],
    iter3: Iterable[_T3],
    iter4: Iterable[_T4],
    iter5: Iterable[_T5],
    iter6: Iterable[_T6],
) -> Iterator[_S]: ...
@overload
def imap(
    func: Callable[..., _S],
    iter1: Iterable[Any],
    iter2: Iterable[Any],
    iter3: Iterable[Any],
    iter4: Iterable[Any],
    iter5: Iterable[Any],
    iter6: Iterable[Any],
    iter7: Iterable[Any],
    *iterables: Iterable[Any],
) -> Iterator[_S]: ...
def starmap(func: Any, iterable: Iterable[Any]) -> Iterator[Any]: ...
def takewhile(predicate: Callable[[_T], Any], iterable: Iterable[_T]) -> Iterator[_T]: ...
def tee(iterable: Iterable[_T], n: int = ...) -> tuple[Iterator[_T], ...]: ...
@overload
def izip(iter1: Iterable[_T1]) -> Iterator[tuple[_T1]]: ...
@overload
def izip(iter1: Iterable[_T1], iter2: Iterable[_T2]) -> Iterator[tuple[_T1, _T2]]: ...
@overload
def izip(iter1: Iterable[_T1], iter2: Iterable[_T2], iter3: Iterable[_T3]) -> Iterator[tuple[_T1, _T2, _T3]]: ...
@overload
def izip(
    iter1: Iterable[_T1], iter2: Iterable[_T2], iter3: Iterable[_T3], iter4: Iterable[_T4]
) -> Iterator[tuple[_T1, _T2, _T3, _T4]]: ...
@overload
def izip(
    iter1: Iterable[_T1], iter2: Iterable[_T2], iter3: Iterable[_T3], iter4: Iterable[_T4], iter5: Iterable[_T5]
) -> Iterator[tuple[_T1, _T2, _T3, _T4, _T5]]: ...
@overload
def izip(
    iter1: Iterable[_T1],
    iter2: Iterable[_T2],
    iter3: Iterable[_T3],
    iter4: Iterable[_T4],
    iter5: Iterable[_T5],
    iter6: Iterable[_T6],
) -> Iterator[tuple[_T1, _T2, _T3, _T4, _T5, _T6]]: ...
@overload
def izip(
    iter1: Iterable[Any],
    iter2: Iterable[Any],
    iter3: Iterable[Any],
    iter4: Iterable[Any],
    iter5: Iterable[Any],
    iter6: Iterable[Any],
    iter7: Iterable[Any],
    *iterables: Iterable[Any],
) -> Iterator[tuple[Any, ...]]: ...
def izip_longest(*p: Iterable[Any], fillvalue: Any = ...) -> Iterator[Any]: ...
@overload
def product(iter1: Iterable[_T1]) -> Iterator[tuple[_T1]]: ...
@overload
def product(iter1: Iterable[_T1], iter2: Iterable[_T2]) -> Iterator[tuple[_T1, _T2]]: ...
@overload
def product(iter1: Iterable[_T1], iter2: Iterable[_T2], iter3: Iterable[_T3]) -> Iterator[tuple[_T1, _T2, _T3]]: ...
@overload
def product(
    iter1: Iterable[_T1], iter2: Iterable[_T2], iter3: Iterable[_T3], iter4: Iterable[_T4]
) -> Iterator[tuple[_T1, _T2, _T3, _T4]]: ...
@overload
def product(
    iter1: Iterable[_T1], iter2: Iterable[_T2], iter3: Iterable[_T3], iter4: Iterable[_T4], iter5: Iterable[_T5]
) -> Iterator[tuple[_T1, _T2, _T3, _T4, _T5]]: ...
@overload
def product(
    iter1: Iterable[_T1],
    iter2: Iterable[_T2],
    iter3: Iterable[_T3],
    iter4: Iterable[_T4],
    iter5: Iterable[_T5],
    iter6: Iterable[_T6],
) -> Iterator[tuple[_T1, _T2, _T3, _T4, _T5, _T6]]: ...
@overload
def product(
    iter1: Iterable[Any],
    iter2: Iterable[Any],
    iter3: Iterable[Any],
    iter4: Iterable[Any],
    iter5: Iterable[Any],
    iter6: Iterable[Any],
    iter7: Iterable[Any],
    *iterables: Iterable[Any],
) -> Iterator[tuple[Any, ...]]: ...
@overload
def product(*iterables: Iterable[Any], repeat: int) -> Iterator[tuple[Any, ...]]: ...
def permutations(iterable: Iterable[_T], r: int = ...) -> Iterator[Sequence[_T]]: ...
def combinations(iterable: Iterable[_T], r: int) -> Iterator[Sequence[_T]]: ...
def combinations_with_replacement(iterable: Iterable[_T], r: int) -> Iterator[Sequence[_T]]: ...



import abc
from typing import Any

class FileDetector(metaclass=abc.ABCMeta):
    __metaclass__: Any
    @abc.abstractmethod
    def is_local_file(self, *keys): ...

class UselessFileDetector(FileDetector):
    def is_local_file(self, *keys) -> None: ...

class LocalFileDetector(FileDetector):
    def is_local_file(self, *keys): ...



"""Generic functions."""

import collections
import contextlib
import itertools
import os.path
import re
import subprocess
import sys
import threading
import traceback
from typing import Iterable, List
import weakref


# We disable the check that keeps pytype from running on not-yet-supported
# versions when we detect that a pytype test is executing, in order to be able
# to test upcoming versions.
def _validate_python_version_upper_bound():
  for frame_summary in traceback.extract_stack():
    head, tail = os.path.split(frame_summary.filename)
    if "/pytype/" in head + "/" and (
        tail.startswith("test_") or tail.endswith("_test.py")):
      return False
  return True


_VALIDATE_PYTHON_VERSION_UPPER_BOUND = _validate_python_version_upper_bound()


def message(error):
  """A convenience function which extracts a message from an exception.

  Use this to replace exception.message, which is deprecated in python2 and
  removed in python3.

  Args:
    error: The exception.

  Returns:
    A message string.
  """
  return error.args[0] if error.args else ""


class UsageError(Exception):
  """Raise this for top-level usage errors."""


def format_version(python_version):
  """Format a version tuple into a dotted version string."""
  return ".".join(str(x) for x in python_version)


def version_from_string(version_string):
  """Parse a version string like "3.7" into a tuple."""
  return tuple(map(int, version_string.split(".")))


def validate_version(python_version):
  """Raise an exception if the python version is unsupported."""
  if len(python_version) != 2:
    # This is typically validated in the option parser, but check here too in
    # case we get python_version via a different entry point.
    raise UsageError("python_version must be <major>.<minor>: %r" %
                     format_version(python_version))
  elif python_version <= (2, 7):
    raise UsageError("Python version %r is not supported. "
                     "Use pytype release 2021.08.03 for Python 2 support." %
                     format_version(python_version))
  elif (2, 8) <= python_version < (3, 0):
    raise UsageError("Python version %r is not a valid Python version." %
                     format_version(python_version))
  elif (3, 0) <= python_version <= (3, 6):
    raise UsageError(
        "Python versions 3.0 - 3.6 are not supported. Use 3.7 and higher.")
  elif python_version > (3, 10) and _VALIDATE_PYTHON_VERSION_UPPER_BOUND:
    # We have an explicit per-minor-version mapping in opcodes.py
    raise UsageError("Python versions > 3.10 are not yet supported.")


def strip_prefix(string, prefix):
  """Strip off prefix if it exists."""
  if string.startswith(prefix):
    return string[len(prefix):]
  return string


def maybe_truncate(s, length=30):
  """Truncate long strings (and append '...'), but leave short strings alone."""
  s = str(s)
  if len(s) > length-3:
    return s[0:length-3] + "..."
  else:
    return s


def pretty_conjunction(conjunction):
  """Pretty-print a conjunction. Use parentheses as necessary.

  E.g. ["a", "b"] -> "(a & b)"

  Args:
    conjunction: List of strings.
  Returns:
    A pretty-printed string.
  """
  if not conjunction:
    return "true"
  elif len(conjunction) == 1:
    return conjunction[0]
  else:
    return "(" + " & ".join(conjunction) + ")"


def pretty_dnf(dnf):
  """Pretty-print a disjunctive normal form (disjunction of conjunctions).

  E.g. [["a", "b"], ["c"]] -> "(a & b) | c".

  Args:
    dnf: A list of list of strings. (Disjunction of conjunctions of strings)
  Returns:
    A pretty-printed string.
  """
  if not dnf:
    return "false"
  else:
    return " | ".join(pretty_conjunction(c) for c in dnf)


def numeric_sort_key(s):
  return tuple((int(e) if e.isdigit() else e) for e in re.split(r"(\d+)", s))


def concat_tuples(tuples):
  return tuple(itertools.chain.from_iterable(tuples))


def native_str(s, errors="strict"):
  """Convert a bytes object to the native str type."""
  if isinstance(s, str):
    return s
  else:
    assert isinstance(s, bytes)
    return s.decode("utf-8", errors)


def get_python_exes(python_version) -> Iterable[List[str]]:
  """Find possible python executables to use.

  Arguments:
    python_version: the version tuple (e.g. (3, 7))
  Yields:
    The path to the executable
  """
  for version in (format_version(python_version), "3"):
    if sys.platform == "win32":
      python_exe = ["py", f"-{version}"]
    else:
      python_exe = [f"python{version}"]
    yield python_exe


def get_python_exe_version(python_exe: List[str]):
  """Determine the major and minor version of given Python executable.

  Arguments:
    python_exe: absolute path to the Python executable
  Returns:
    Version as (major, minor) tuple.
  """
  try:
    python_exe_version = subprocess.check_output(
        python_exe + ["-V"], stderr=subprocess.STDOUT).decode()
  except (subprocess.CalledProcessError, FileNotFoundError):
    return None

  return parse_exe_version_string(python_exe_version)


def parse_exe_version_string(version_str):
  """Parse the version string of a Python executable.

  Arguments:
    version_str: Version string as emitted by running `PYTHON_EXE -V`
  Returns:
    Version as (major, minor) tuple.
  """
  # match the major.minor part of the version string, ignore the micro part
  matcher = re.search(r"Python (\d+\.\d+)\.\d+", version_str)

  if matcher:
    return version_from_string(matcher.group(1))
  else:
    return None


def can_compile_bytecode_natively(python_version):
  # Optimization: calling compile_bytecode directly is faster than spawning a
  # subprocess and lets us avoid extracting a large Python executable into /tmp.
  # We can do this only when the host and target versions match.
  return python_version == sys.version_info[:2]


def list_startswith(l, prefix):
  """Like str.startswith, but for lists."""
  return l[:len(prefix)] == prefix


def list_strip_prefix(l, prefix):
  """Remove prefix, if it's there."""
  return l[len(prefix):] if list_startswith(l, prefix) else l


def _arg_names(f):
  """Return the argument names of a function."""
  return f.__code__.co_varnames[:f.__code__.co_argcount]


def invert_dict(d):
  """Invert a dictionary.

  Converts a dictionary (mapping strings to lists of strings) to a dictionary
  that maps into the other direction.

  Arguments:
    d: Dictionary to be inverted

  Returns:
    A dictionary n with the property that if "y in d[x]", then "x in n[y]".
  """

  inverted = collections.defaultdict(list)
  for key, value_list in d.items():
    for val in value_list:
      inverted[val].append(key)
  return inverted


def unique_list(xs):
  """Return a unique list from an iterable, preserving order."""
  seen = set()
  out = []
  for x in xs:
    if x not in seen:
      seen.add(x)
      out.append(x)
  return out


class DynamicVar:
  """A dynamically scoped variable.

  This is a per-thread dynamic variable, with an initial value of None.
  The bind() call establishes a new value that will be in effect for the
  duration of the resulting context manager.  This is intended to be used
  in conjunction with a decorator.
  """

  def __init__(self):
    self._local = threading.local()

  def _values(self):
    values = getattr(self._local, "values", None)
    if values is None:
      values = [None]  # Stack of bindings, with an initial default of None.
      self._local.values = values
    return values

  @contextlib.contextmanager
  def bind(self, value):
    """Bind the dynamic variable to the supplied value."""
    values = self._values()
    try:
      values.append(value)  # Push the new binding.
      yield
    finally:
      values.pop()  # Pop the binding.

  def get(self):
    """Return the current value of the dynamic variable."""
    return self._values()[-1]


class AnnotatingDecorator:
  """A decorator for storing function attributes.

  Attributes:
    lookup: maps functions to their attributes.
  """

  def __init__(self):
    self.lookup = {}

  def __call__(self, value):
    def decorate(f):
      self.lookup[f.__name__] = value
      return f
    return decorate


class ContextWeakrefMixin:

  __slots__ = ["ctx_weakref"]

  def __init__(self, ctx):
    self.ctx_weakref = weakref.ref(ctx)

  @property
  def ctx(self):
    return self.ctx_weakref()



from typing import Any

from openpyxl.descriptors.serialisable import Serialisable

class CustomChartsheetView(Serialisable):
    tagname: str
    guid: Any
    scale: Any
    state: Any
    zoomToFit: Any
    pageMargins: Any
    pageSetup: Any
    headerFooter: Any
    __elements__: Any
    def __init__(
        self,
        guid: Any | None = ...,
        scale: Any | None = ...,
        state: str = ...,
        zoomToFit: Any | None = ...,
        pageMargins: Any | None = ...,
        pageSetup: Any | None = ...,
        headerFooter: Any | None = ...,
    ) -> None: ...

class CustomChartsheetViews(Serialisable):
    tagname: str
    customSheetView: Any
    __elements__: Any
    def __init__(self, customSheetView: Any | None = ...) -> None: ...



from typing import Any

import sqlalchemy.types as sqltypes

from ...sql import functions as sqlfunc

class HSTORE(sqltypes.Indexable, sqltypes.Concatenable, sqltypes.TypeEngine):
    __visit_name__: str
    hashable: bool
    text_type: Any
    def __init__(self, text_type: Any | None = ...) -> None: ...

    class Comparator(sqltypes.Indexable.Comparator, sqltypes.Concatenable.Comparator):
        def has_key(self, other): ...
        def has_all(self, other): ...
        def has_any(self, other): ...
        def contains(self, other, **kwargs): ...
        def contained_by(self, other): ...
        def defined(self, key): ...
        def delete(self, key): ...
        def slice(self, array): ...
        def keys(self): ...
        def vals(self): ...
        def array(self): ...
        def matrix(self): ...
    comparator_factory: Any
    def bind_processor(self, dialect): ...
    def result_processor(self, dialect, coltype): ...

class hstore(sqlfunc.GenericFunction):
    type: Any
    name: str
    inherit_cache: bool

class _HStoreDefinedFunction(sqlfunc.GenericFunction):
    type: Any
    name: str
    inherit_cache: bool

class _HStoreDeleteFunction(sqlfunc.GenericFunction):
    type: Any
    name: str
    inherit_cache: bool

class _HStoreSliceFunction(sqlfunc.GenericFunction):
    type: Any
    name: str
    inherit_cache: bool

class _HStoreKeysFunction(sqlfunc.GenericFunction):
    type: Any
    name: str
    inherit_cache: bool

class _HStoreValsFunction(sqlfunc.GenericFunction):
    type: Any
    name: str
    inherit_cache: bool

class _HStoreArrayFunction(sqlfunc.GenericFunction):
    type: Any
    name: str
    inherit_cache: bool

class _HStoreMatrixFunction(sqlfunc.GenericFunction):
    type: Any
    name: str
    inherit_cache: bool



# Utility types for typeshed

# This module contains various common types to be used by typeshed. The
# module and its types do not exist at runtime. You can use this module
# outside of typeshed, but no API stability guarantees are made. To use
# it in implementation (.py) files, the following construct must be used:
#
#     from typing import TYPE_CHECKING
#     if TYPE_CHECKING:
#         from _typeshed import ...
#
# If on Python versions < 3.10 and "from __future__ import annotations"
# is not used, types from this module must be quoted.

import array
import mmap
from typing import Any, Container, Iterable, Protocol, Text, TypeVar
from typing_extensions import Literal, final

_KT = TypeVar("_KT")
_KT_co = TypeVar("_KT_co", covariant=True)
_KT_contra = TypeVar("_KT_contra", contravariant=True)
_VT = TypeVar("_VT")
_VT_co = TypeVar("_VT_co", covariant=True)
_T = TypeVar("_T")
_T_co = TypeVar("_T_co", covariant=True)
_T_contra = TypeVar("_T_contra", contravariant=True)

# Use for "self" annotations:
#   def __enter__(self: Self) -> Self: ...
Self = TypeVar("Self")  # noqa: Y001

class IdentityFunction(Protocol):
    def __call__(self, __x: _T) -> _T: ...

class SupportsLessThan(Protocol):
    def __lt__(self, __other: Any) -> bool: ...

SupportsLessThanT = TypeVar("SupportsLessThanT", bound=SupportsLessThan)  # noqa: Y001

class SupportsDivMod(Protocol[_T_contra, _T_co]):
    def __divmod__(self, __other: _T_contra) -> _T_co: ...

class SupportsRDivMod(Protocol[_T_contra, _T_co]):
    def __rdivmod__(self, __other: _T_contra) -> _T_co: ...

# Mapping-like protocols

class SupportsItems(Protocol[_KT_co, _VT_co]):
    # We want dictionaries to support this on Python 2.
    def items(self) -> Iterable[tuple[_KT_co, _VT_co]]: ...

class SupportsKeysAndGetItem(Protocol[_KT, _VT_co]):
    def keys(self) -> Iterable[_KT]: ...
    def __getitem__(self, __k: _KT) -> _VT_co: ...

class SupportsGetItem(Container[_KT_contra], Protocol[_KT_contra, _VT_co]):
    def __getitem__(self, __k: _KT_contra) -> _VT_co: ...

class SupportsItemAccess(SupportsGetItem[_KT_contra, _VT], Protocol[_KT_contra, _VT]):
    def __setitem__(self, __k: _KT_contra, __v: _VT) -> None: ...
    def __delitem__(self, __v: _KT_contra) -> None: ...

# These aliases can be used in places where a PathLike object can be used
# instead of a string in Python 3.
StrPath = Text
BytesPath = str
StrOrBytesPath = Text
AnyPath = StrOrBytesPath  # obsolete, will be removed soon

OpenTextModeUpdating = Literal[
    "r+",
    "+r",
    "rt+",
    "r+t",
    "+rt",
    "tr+",
    "t+r",
    "+tr",
    "w+",
    "+w",
    "wt+",
    "w+t",
    "+wt",
    "tw+",
    "t+w",
    "+tw",
    "a+",
    "+a",
    "at+",
    "a+t",
    "+at",
    "ta+",
    "t+a",
    "+ta",
    "x+",
    "+x",
    "xt+",
    "x+t",
    "+xt",
    "tx+",
    "t+x",
    "+tx",
]
OpenTextModeWriting = Literal["w", "wt", "tw", "a", "at", "ta", "x", "xt", "tx"]
OpenTextModeReading = Literal["r", "rt", "tr", "U", "rU", "Ur", "rtU", "rUt", "Urt", "trU", "tUr", "Utr"]
OpenTextMode = OpenTextModeUpdating | OpenTextModeWriting | OpenTextModeReading
OpenBinaryModeUpdating = Literal[
    "rb+",
    "r+b",
    "+rb",
    "br+",
    "b+r",
    "+br",
    "wb+",
    "w+b",
    "+wb",
    "bw+",
    "b+w",
    "+bw",
    "ab+",
    "a+b",
    "+ab",
    "ba+",
    "b+a",
    "+ba",
    "xb+",
    "x+b",
    "+xb",
    "bx+",
    "b+x",
    "+bx",
]
OpenBinaryModeWriting = Literal["wb", "bw", "ab", "ba", "xb", "bx"]
OpenBinaryModeReading = Literal["rb", "br", "rbU", "rUb", "Urb", "brU", "bUr", "Ubr"]
OpenBinaryMode = OpenBinaryModeUpdating | OpenBinaryModeReading | OpenBinaryModeWriting

class HasFileno(Protocol):
    def fileno(self) -> int: ...

FileDescriptor = int
FileDescriptorLike = int | HasFileno

class SupportsRead(Protocol[_T_co]):
    def read(self, __length: int = ...) -> _T_co: ...

class SupportsReadline(Protocol[_T_co]):
    def readline(self, __length: int = ...) -> _T_co: ...

class SupportsNoArgReadline(Protocol[_T_co]):
    def readline(self) -> _T_co: ...

class SupportsWrite(Protocol[_T_contra]):
    def write(self, __s: _T_contra) -> Any: ...

ReadableBuffer = bytes | bytearray | memoryview | array.array[Any] | mmap.mmap | buffer
WriteableBuffer = bytearray | memoryview | array.array[Any] | mmap.mmap | buffer

# Used by type checkers for checks involving None (does not exist at runtime)
@final
class NoneType:
    def __bool__(self) -> Literal[False]: ...



"""Tests for tool_utils.py."""

import os

from pytype import file_utils
from pytype.tools import tool_utils
import unittest


class TestSetupLoggingOrDie(unittest.TestCase):
  """Tests for tool_utils.setup_logging_or_die."""

  def test_negative_verbosity(self):
    with self.assertRaises(SystemExit):
      tool_utils.setup_logging_or_die(-1)

  def test_excessive_verbosity(self):
    with self.assertRaises(SystemExit):
      tool_utils.setup_logging_or_die(3)

  def test_set_level(self):
    # Log level can't actually be set in a test, so we're just testing that the
    # code doesn't blow up.
    tool_utils.setup_logging_or_die(0)
    tool_utils.setup_logging_or_die(1)
    tool_utils.setup_logging_or_die(2)


class TestMakeDirsOrDie(unittest.TestCase):
  """Tests for tool_utils.makedirs_or_die()."""

  def test_make(self):
    with file_utils.Tempdir() as d:
      subdir = os.path.join(d.path, 'some/path')
      tool_utils.makedirs_or_die(subdir, '')
      self.assertTrue(os.path.isdir(subdir))

  def test_die(self):
    with self.assertRaises(SystemExit):
      tool_utils.makedirs_or_die('/nonexistent/path', '')


if __name__ == '__main__':
  unittest.main()



import ctypes
from collections.abc import Callable, Iterable, Sequence
from ctypes import _CData, _SimpleCData, c_char
from multiprocessing.context import BaseContext
from multiprocessing.synchronize import _LockLike
from types import TracebackType
from typing import Any, Generic, Protocol, TypeVar, overload
from typing_extensions import Literal

__all__ = ["RawValue", "RawArray", "Value", "Array", "copy", "synchronized"]

_T = TypeVar("_T")
_CT = TypeVar("_CT", bound=_CData)

@overload
def RawValue(typecode_or_type: type[_CT], *args: Any) -> _CT: ...
@overload
def RawValue(typecode_or_type: str, *args: Any) -> Any: ...
@overload
def RawArray(typecode_or_type: type[_CT], size_or_initializer: int | Sequence[Any]) -> ctypes.Array[_CT]: ...
@overload
def RawArray(typecode_or_type: str, size_or_initializer: int | Sequence[Any]) -> Any: ...
@overload
def Value(typecode_or_type: type[_CT], *args: Any, lock: Literal[False], ctx: BaseContext | None = ...) -> _CT: ...
@overload
def Value(
    typecode_or_type: type[_CT], *args: Any, lock: Literal[True] | _LockLike, ctx: BaseContext | None = ...
) -> SynchronizedBase[_CT]: ...
@overload
def Value(
    typecode_or_type: str, *args: Any, lock: Literal[True] | _LockLike, ctx: BaseContext | None = ...
) -> SynchronizedBase[Any]: ...
@overload
def Value(
    typecode_or_type: str | type[_CData], *args: Any, lock: bool | _LockLike = ..., ctx: BaseContext | None = ...
) -> Any: ...
@overload
def Array(
    typecode_or_type: type[_CT], size_or_initializer: int | Sequence[Any], *, lock: Literal[False], ctx: BaseContext | None = ...
) -> _CT: ...
@overload
def Array(
    typecode_or_type: type[_CT],
    size_or_initializer: int | Sequence[Any],
    *,
    lock: Literal[True] | _LockLike,
    ctx: BaseContext | None = ...,
) -> SynchronizedArray[_CT]: ...
@overload
def Array(
    typecode_or_type: str,
    size_or_initializer: int | Sequence[Any],
    *,
    lock: Literal[True] | _LockLike,
    ctx: BaseContext | None = ...,
) -> SynchronizedArray[Any]: ...
@overload
def Array(
    typecode_or_type: str | type[_CData],
    size_or_initializer: int | Sequence[Any],
    *,
    lock: bool | _LockLike = ...,
    ctx: BaseContext | None = ...,
) -> Any: ...
def copy(obj: _CT) -> _CT: ...
@overload
def synchronized(obj: _SimpleCData[_T], lock: _LockLike | None = ..., ctx: Any | None = ...) -> Synchronized[_T]: ...
@overload
def synchronized(obj: ctypes.Array[c_char], lock: _LockLike | None = ..., ctx: Any | None = ...) -> SynchronizedString: ...
@overload
def synchronized(obj: ctypes.Array[_CT], lock: _LockLike | None = ..., ctx: Any | None = ...) -> SynchronizedArray[_CT]: ...
@overload
def synchronized(obj: _CT, lock: _LockLike | None = ..., ctx: Any | None = ...) -> SynchronizedBase[_CT]: ...

class _AcquireFunc(Protocol):
    def __call__(self, block: bool = ..., timeout: float | None = ...) -> bool: ...

class SynchronizedBase(Generic[_CT]):
    acquire: _AcquireFunc
    release: Callable[[], None]
    def __init__(self, obj: Any, lock: _LockLike | None = ..., ctx: Any | None = ...) -> None: ...
    def __reduce__(self) -> tuple[Callable[[Any, _LockLike], SynchronizedBase[Any]], tuple[Any, _LockLike]]: ...
    def get_obj(self) -> _CT: ...
    def get_lock(self) -> _LockLike: ...
    def __enter__(self) -> bool: ...
    def __exit__(
        self, __exc_type: type[BaseException] | None, __exc_val: BaseException | None, __exc_tb: TracebackType | None
    ) -> None: ...

class Synchronized(SynchronizedBase[_SimpleCData[_T]], Generic[_T]):
    value: _T

class SynchronizedArray(SynchronizedBase[ctypes.Array[_CT]], Generic[_CT]):
    def __len__(self) -> int: ...
    def __getitem__(self, i: int) -> _CT: ...
    def __setitem__(self, i: int, value: _CT) -> None: ...
    def __getslice__(self, start: int, stop: int) -> list[_CT]: ...
    def __setslice__(self, start: int, stop: int, values: Iterable[_CT]) -> None: ...

class SynchronizedString(SynchronizedArray[c_char]):
    value: bytes
    raw: bytes



"""Classes for instrumenting code to collect various metrics.

Instrumentation consists of creating the metric and then updating it.
Creation can be performed at the module level or as a class attribute.  Since
the metric namespace is global, metrics should not be created by instances
unless that instance is certain to be a singleton.

Sample code:

_my_counter = metrics.Counter("my-counter")


def foo():
  _my_counter.inc()  # calls to foo() count as 1 unit.


def bar(n):
  _my_counter.inc(n)  # calls to bar() count as n units.
"""

import json
import math
import re
import time
import tracemalloc


# Metric serialization/deserialization code, taking advantage of the fact that
# all instance vars of metrics are things that json can serialize, so we don't
# need to write custom JsonEncoder and JsonDecoder classes per Metric subclass.

# Register metric types for deserialization
_METRIC_TYPES = {}


class _RegistryMeta(type):
  """Metaclass that registers subclasses in _METRIC_TYPES."""

  def __new__(cls, name, bases, class_dict):
    subcls = super().__new__(cls, name, bases, class_dict)
    _METRIC_TYPES[subcls.__name__] = subcls
    return subcls


def _deserialize(typ, payload):
  """Construct a Metric from a typename and payload loaded from json."""
  if typ not in _METRIC_TYPES:
    raise TypeError(f"Could not decode class {typ}")
  cls = _METRIC_TYPES[typ]
  out = cls(None)
  out.__dict__.update(payload)
  return out


def _serialize(obj):
  """Return a json-serializable form of object."""
  return [obj.__class__.__name__, vars(obj)]


def dump_all(objs, fp):
  """Write a list of metrics to a json file."""
  json.dump([_serialize(x) for x in objs], fp)


def load_all(fp):
  """Read a list of metrics from a json file."""
  metrics = json.load(fp)
  return [_deserialize(*x) for x in metrics]


_METRIC_NAME_RE = re.compile(r"^[a-zA-Z_]\w+$")

_registered_metrics = {}  # Map from metric name to Metric object.
_enabled = False  # True iff metrics should be collected.


def _validate_metric_name(name):
  if _METRIC_NAME_RE.match(name) is None:
    raise ValueError(f"Illegal metric name: {name}")


def _prepare_for_test(enabled=True):
  """Setup metrics collection for a test."""
  _registered_metrics.clear()
  global _enabled
  _enabled = enabled


def get_cpu_clock():
  """Returns CPU clock to keep compatibility with various Python versions."""
  return time.process_time()


def get_metric(name, constructor, *args, **kwargs):
  """Return an existing metric or create a new one for the given name.

  Args:
    name: The name of the metric.
    constructor: A class to instantiate if a new metric is required.
    *args: Additional positional args to pass to the constructor.
    **kwargs: Keyword args for the constructor.

  Returns:
    The current metric registered to name, or a new one created by
    invoking constructor(name, *args, **kwargs).
  """
  metric = _registered_metrics.get(name)
  if metric is not None:
    return metric
  else:
    return constructor(name, *args, **kwargs)


def get_report():
  """Return a string listing all metrics, one per line."""
  lines = [str(_registered_metrics[n]) + "\n"
           for n in sorted(_registered_metrics)]
  return "".join(lines)


def merge_from_file(metrics_file):
  """Merge metrics recorded in another file into the current metrics."""
  for metric in load_all(metrics_file):
    existing = _registered_metrics.get(metric.name)
    if existing is None:
      _validate_metric_name(metric.name)
      _registered_metrics[metric.name] = metric
    else:
      if type(metric) != type(existing):  # pylint: disable=unidiomatic-typecheck
        raise TypeError("Cannot merge metrics of different types.")
      existing._merge(metric)  # pylint: disable=protected-access


class Metric(metaclass=_RegistryMeta):
  """Abstract base class for metrics."""

  def __init__(self, name):
    """Initialize the metric and register it under the specified name."""
    if name is None:
      # We do not want to register this metric (e.g. we are deserializing a
      # metric from file and need to merge it into the existing metric with the
      # same name.)
      return
    _validate_metric_name(name)
    if name in _registered_metrics:
      raise ValueError(f"Metric {name} has already been defined.")
    self._name = name
    _registered_metrics[name] = self

  @property
  def name(self):
    return self._name

  def _summary(self):
    """Return a string summarizing the value of the metric."""
    raise NotImplementedError

  def _merge(self, other):
    """Merge data from another metric of the same type."""
    raise NotImplementedError

  def __str__(self):
    return f"{self._name}: {self._summary()}"


class Counter(Metric):
  """A monotonically increasing metric."""

  def __init__(self, name):
    super().__init__(name)
    self._total = 0

  def inc(self, count=1):
    """Increment the metric by the specified amount."""
    if count < 0:
      raise ValueError("Counter must be monotonically increasing.")
    if not _enabled:
      return
    self._total += count

  def _summary(self):
    return str(self._total)

  def _merge(self, other):
    # pylint: disable=protected-access
    self._total += other._total


class StopWatch(Metric):
  """A counter that measures the time spent in a "with" statement."""

  def __enter__(self):
    self._start_time = get_cpu_clock()

  def __exit__(self, exc_type, exc_value, traceback):
    self._total = get_cpu_clock() - self._start_time
    del self._start_time

  def _summary(self):
    return f"{self._total:f} seconds"

  def _merge(self, other):
    # pylint: disable=protected-access
    self._total += other._total


class ReentrantStopWatch(Metric):
  """A watch that supports being called multiple times and recursively."""

  def __init__(self, name):
    super().__init__(name)
    self._time = 0
    self._calls = 0

  def __enter__(self):
    if not self._calls:
      self._start_time = get_cpu_clock()
    self._calls += 1

  def __exit__(self, exc_type, exc_value, traceback):
    self._calls -= 1
    if not self._calls:
      self._time += get_cpu_clock() - self._start_time
      del self._start_time

  def _merge(self, other):
    self._time += other._time  # pylint: disable=protected-access

  def _summary(self):
    return f"time spend below this StopWatch: {self._time}"


class MapCounter(Metric):
  """A set of related counters keyed by an arbitrary string."""

  def __init__(self, name):
    super().__init__(name)
    self._counts = {}
    self._total = 0

  def inc(self, key, count=1):
    """Increment the metric by the specified amount.

    Args:
      key: A string to be used as the key.
      count: The amount to increment by (non-negative integer).

    Raises:
      ValueError: if the count is less than 0.
    """
    if count < 0:
      raise ValueError("Counter must be monotonically increasing.")
    if not _enabled:
      return
    self._counts[key] = self._counts.get(key, 0) + count
    self._total += count

  def _summary(self):
    details = ", ".join(["%s=%d" % (k, self._counts[k])
                         for k in sorted(self._counts)])
    return "%d {%s}" % (self._total, details)

  def _merge(self, other):
    # pylint: disable=protected-access
    for key, count in other._counts.items():
      self._counts[key] = self._counts.get(key, 0) + count
      self._total += count


class Distribution(Metric):
  """A metric to track simple statistics from a distribution of values."""

  def __init__(self, name):
    super().__init__(name)
    self._count = 0  # Number of values.
    self._total = 0.0  # Sum of the values.
    self._squared = 0.0  # Sum of the squares of the values.
    self._min = None
    self._max = None

  def add(self, value):
    """Add a value to the distribution."""
    if not _enabled:
      return
    self._count += 1
    self._total += value
    self._squared += value * value
    if self._min is None:
      # First add, this value is the min and max
      self._min = self._max = value
    else:
      self._min = min(self._min, value)
      self._max = max(self._max, value)

  def _mean(self):
    if self._count:
      return self._total / float(self._count)

  def _stdev(self):
    if self._count:
      variance = ((self._squared * self._count - self._total * self._total) /
                  (self._count * self._count))
      if variance < 0.0:
        # This can only happen as the result of rounding error when the actual
        # variance is very, very close to 0.  Assume it is 0.
        return 0.0
      return  math.sqrt(variance)

  def _summary(self):
    return "total=%s, count=%d, min=%s, max=%s, mean=%s, stdev=%s" % (
        self._total, self._count, self._min, self._max, self._mean(),
        self._stdev())

  def _merge(self, other):
    # pylint: disable=protected-access
    if other._count == 0:
      # Exit early so we don't have to worry about min/max of None.
      return
    self._count += other._count
    self._total += other._total
    self._squared += other._squared
    if self._min is None:
      self._min = other._min
      self._max = other._max
    else:
      self._min = min(self._min, other._min)
      self._max = max(self._max, other._max)


class Snapshot(Metric):
  """A metric to track memory usage via tracemalloc snapshots."""

  def __init__(self, name, enabled=False, groupby="lineno",
               nframes=1, count=10):
    super().__init__(name)
    self.snapshots = []
    # The metric to group memory blocks by. Default is "lineno", which groups by
    # which file and line allocated the block. The other useful value is
    # "traceback", which groups by the stack frames leading to each allocation.
    self.groupby = groupby
    # The number of stack frames to store per memory block. Values greater than
    # 1 are only useful if groupby = "traceback".
    self.nframes = nframes
    # The number of memory block statistics to save.
    self.count = count
    self.running = False
    # Two conditions must be met for memory snapshots to be taken:
    # 1. Metrics have been enabled (global _enabled)
    # 2. Explicitly enabled by the arg to the constructor (which should be the
    # options.memory_snapshot flag set by the --memory-snapshots option)
    self.enabled = _enabled and enabled

  def _start_tracemalloc(self):
    tracemalloc.start(self.nframes)
    self.running = True

  def _stop_tracemalloc(self):
    tracemalloc.stop()
    self.running = False

  def take_snapshot(self, where=""):
    """Stores a tracemalloc snapshot."""
    if not self.enabled:
      return
    if not self.running:
      self._start_tracemalloc()
    snap = tracemalloc.take_snapshot()
    # Store the top self.count memory consumers by self.groupby
    # We can't just store the list of statistics though! Statistic.__eq__
    # doesn't take None into account during comparisons, and json will compare
    # it to None when trying to process it, causing an error. So, store it as a
    # string instead.
    self.snapshots.append("{}:\n{}".format(where, "\n".join(
        map(str, snap.statistics(self.groupby)[:self.count]))))

  def __enter__(self):
    if not self.enabled:
      return
    self._start_tracemalloc()
    self.take_snapshot("__enter__")

  def __exit__(self, exc_type, exc_value, traceback):
    if not self.running:
      return
    self.take_snapshot("__exit__")
    self._stop_tracemalloc()

  def _summary(self):
    return "\n\n".join(self.snapshots)


class MetricsContext:
  """A context manager that configures metrics and writes their output."""

  def __init__(self, output_path, open_function=open):
    """Initialize.

    Args:
      output_path: The path for the metrics data.  If empty, no metrics are
          collected.
      open_function: A custom file opening function.
    """
    self._output_path = output_path
    self._open_function = open_function
    self._old_enabled = None  # Set in __enter__.

  def __enter__(self):
    global _enabled
    self._old_enabled = _enabled
    _enabled = bool(self._output_path)

  def __exit__(self, exc_type, exc_value, traceback):
    global _enabled
    _enabled = self._old_enabled
    if self._output_path:
      with self._open_function(self._output_path, "w") as f:
        dump_all(_registered_metrics.values(), f)



from typing import Any

from .models.segment import SegmentContextManager as SegmentContextManager
from .models.subsegment import (
    SubsegmentContextManager as SubsegmentContextManager,
    is_already_recording as is_already_recording,
    subsegment_decorator as subsegment_decorator,
)
from .recorder import AWSXRayRecorder as AWSXRayRecorder
from .utils import stacktrace as stacktrace

class AsyncSegmentContextManager(SegmentContextManager):
    async def __aenter__(self): ...
    async def __aexit__(self, exc_type, exc_val, exc_tb): ...

class AsyncSubsegmentContextManager(SubsegmentContextManager):
    async def __call__(self, wrapped, instance, args, kwargs): ...
    async def __aenter__(self): ...
    async def __aexit__(self, exc_type, exc_val, exc_tb): ...

class AsyncAWSXRayRecorder(AWSXRayRecorder):
    def capture_async(self, name: Any | None = ...): ...
    def in_segment_async(self, name: Any | None = ..., **segment_kwargs): ...
    def in_subsegment_async(self, name: Any | None = ..., **subsegment_kwargs): ...
    async def record_subsegment_async(self, wrapped, instance, args, kwargs, name, namespace, meta_processor): ...



from typing import Any, ClassVar

import passlib.utils.handlers as uh

class fshp(uh.HasRounds, uh.HasRawSalt, uh.HasRawChecksum, uh.GenericHandler):  # type: ignore[misc]
    name: ClassVar[str]
    checksum_chars: ClassVar[str]
    ident: ClassVar[str]
    default_salt_size: ClassVar[int]
    max_salt_size: ClassVar[None]
    default_rounds: ClassVar[int]
    min_rounds: ClassVar[int]
    max_rounds: ClassVar[int]
    rounds_cost: ClassVar[str]
    default_variant: ClassVar[int]
    @classmethod
    def using(cls, variant: int | str | bytes | None = ..., **kwds): ...  # type: ignore[override]
    variant: int | None
    use_defaults: Any
    def __init__(self, variant: int | str | bytes | None = ..., **kwds) -> None: ...
    @property
    def checksum_alg(self): ...
    @property
    def checksum_size(self): ...
    @classmethod
    def from_string(cls, hash): ...



"""Utils for creating importlab environment."""

from importlab import environment
from importlab import fs

from pytype import utils


class PytdFileSystem(fs.ExtensionRemappingFileSystem):
  """File system that remaps .py file extensions to pytd."""

  def __init__(self, underlying):
    super().__init__(underlying, 'pytd')


def create_importlab_environment(conf, typeshed):
  """Create an importlab environment from the python version and path."""
  python_version = utils.version_from_string(conf.python_version)
  path = fs.Path()
  for p in conf.pythonpath:
    path.add_path(p, 'os')
  for p in typeshed.get_pytd_paths():
    path.add_fs(PytdFileSystem(fs.OSFileSystem(p)))
  for p in typeshed.get_typeshed_paths():
    path.add_path(p, 'pyi')
  return environment.Environment(path, python_version)



from typing import Any, Text, TypeVar

from google.protobuf.descriptor_pool import DescriptorPool
from google.protobuf.message import Message

_MessageT = TypeVar("_MessageT", bound=Message)

class Error(Exception): ...
class ParseError(Error): ...
class SerializeToJsonError(Error): ...

def MessageToJson(
    message: Message,
    including_default_value_fields: bool = ...,
    preserving_proto_field_name: bool = ...,
    indent: int = ...,
    sort_keys: bool = ...,
    use_integers_for_enums: bool = ...,
    descriptor_pool: DescriptorPool | None = ...,
    float_precision: int | None = ...,
) -> str: ...
def MessageToDict(
    message: Message,
    including_default_value_fields: bool = ...,
    preserving_proto_field_name: bool = ...,
    use_integers_for_enums: bool = ...,
    descriptor_pool: DescriptorPool | None = ...,
    float_precision: int | None = ...,
) -> dict[Text, Any]: ...
def Parse(
    text: bytes | Text, message: _MessageT, ignore_unknown_fields: bool = ..., descriptor_pool: DescriptorPool | None = ...
) -> _MessageT: ...
def ParseDict(
    js_dict: Any, message: _MessageT, ignore_unknown_fields: bool = ..., descriptor_pool: DescriptorPool | None = ...
) -> _MessageT: ...



from typing import Any

from stripe.api_resources import ApplicationFee as ApplicationFee
from stripe.api_resources.abstract import UpdateableAPIResource as UpdateableAPIResource

class ApplicationFeeRefund(UpdateableAPIResource):
    OBJECT_NAME: str
    @classmethod
    def modify(cls, fee, sid, **params): ...
    def instance_url(self): ...
    @classmethod
    def retrieve(cls, id, api_key: Any | None = ..., **params) -> None: ...



from typing import Any

from openpyxl.descriptors.serialisable import Serialisable

class SortCondition(Serialisable):
    tagname: str
    descending: Any
    sortBy: Any
    ref: Any
    customList: Any
    dxfId: Any
    iconSet: Any
    iconId: Any
    def __init__(
        self,
        ref: Any | None = ...,
        descending: Any | None = ...,
        sortBy: Any | None = ...,
        customList: Any | None = ...,
        dxfId: Any | None = ...,
        iconSet: Any | None = ...,
        iconId: Any | None = ...,
    ) -> None: ...

class SortState(Serialisable):
    tagname: str
    columnSort: Any
    caseSensitive: Any
    sortMethod: Any
    ref: Any
    sortCondition: Any
    extLst: Any
    __elements__: Any
    def __init__(
        self,
        columnSort: Any | None = ...,
        caseSensitive: Any | None = ...,
        sortMethod: Any | None = ...,
        ref: Any | None = ...,
        sortCondition=...,
        extLst: Any | None = ...,
    ) -> None: ...
    def __bool__(self): ...

class IconFilter(Serialisable):
    tagname: str
    iconSet: Any
    iconId: Any
    def __init__(self, iconSet: Any | None = ..., iconId: Any | None = ...) -> None: ...

class ColorFilter(Serialisable):
    tagname: str
    dxfId: Any
    cellColor: Any
    def __init__(self, dxfId: Any | None = ..., cellColor: Any | None = ...) -> None: ...

class DynamicFilter(Serialisable):
    tagname: str
    type: Any
    val: Any
    valIso: Any
    maxVal: Any
    maxValIso: Any
    def __init__(
        self,
        type: Any | None = ...,
        val: Any | None = ...,
        valIso: Any | None = ...,
        maxVal: Any | None = ...,
        maxValIso: Any | None = ...,
    ) -> None: ...

class CustomFilter(Serialisable):
    tagname: str
    operator: Any
    val: Any
    def __init__(self, operator: Any | None = ..., val: Any | None = ...) -> None: ...

class CustomFilters(Serialisable):
    tagname: str
    customFilter: Any
    __elements__: Any
    def __init__(self, _and: Any | None = ..., customFilter=...) -> None: ...

class Top10(Serialisable):
    tagname: str
    top: Any
    percent: Any
    val: Any
    filterVal: Any
    def __init__(
        self, top: Any | None = ..., percent: Any | None = ..., val: Any | None = ..., filterVal: Any | None = ...
    ) -> None: ...

class DateGroupItem(Serialisable):
    tagname: str
    year: Any
    month: Any
    day: Any
    hour: Any
    minute: Any
    second: Any
    dateTimeGrouping: Any
    def __init__(
        self,
        year: Any | None = ...,
        month: Any | None = ...,
        day: Any | None = ...,
        hour: Any | None = ...,
        minute: Any | None = ...,
        second: Any | None = ...,
        dateTimeGrouping: Any | None = ...,
    ) -> None: ...

class Filters(Serialisable):
    tagname: str
    blank: Any
    calendarType: Any
    filter: Any
    dateGroupItem: Any
    __elements__: Any
    def __init__(self, blank: Any | None = ..., calendarType: Any | None = ..., filter=..., dateGroupItem=...) -> None: ...

class FilterColumn(Serialisable):
    tagname: str
    colId: Any
    col_id: Any
    hiddenButton: Any
    showButton: Any
    filters: Any
    top10: Any
    customFilters: Any
    dynamicFilter: Any
    colorFilter: Any
    iconFilter: Any
    extLst: Any
    __elements__: Any
    def __init__(
        self,
        colId: Any | None = ...,
        hiddenButton: Any | None = ...,
        showButton: Any | None = ...,
        filters: Any | None = ...,
        top10: Any | None = ...,
        customFilters: Any | None = ...,
        dynamicFilter: Any | None = ...,
        colorFilter: Any | None = ...,
        iconFilter: Any | None = ...,
        extLst: Any | None = ...,
        blank: Any | None = ...,
        vals: Any | None = ...,
    ) -> None: ...

class AutoFilter(Serialisable):
    tagname: str
    ref: Any
    filterColumn: Any
    sortState: Any
    extLst: Any
    __elements__: Any
    def __init__(
        self, ref: Any | None = ..., filterColumn=..., sortState: Any | None = ..., extLst: Any | None = ...
    ) -> None: ...
    def __bool__(self): ...
    def add_filter_column(self, col_id, vals, blank: bool = ...) -> None: ...
    def add_sort_condition(self, ref, descending: bool = ...) -> None: ...



"""Debugging utilities."""

# pylint: disable=protected-access


def dump(node, ast, annotate_fields=True,
         include_attributes=True, indent="  "):
  """Return a formatted dump of the tree in *node*.

  This is mainly useful for debugging purposes.  The returned string will show
  the names and the values for fields.  This makes the code impossible to
  evaluate, so if evaluation is wanted *annotate_fields* must be set to False.
  Attributes such as line numbers and column offsets are dumped by default. If
  this is not wanted, *include_attributes* can be set to False.

  Arguments:
    node: Top AST node.
    ast: An module providing an AST class hierarchy.
    annotate_fields: Show field annotations.
    include_attributes: Show all attributes.
    indent: Indentation string.

  Returns:
    A formatted tree.
  """
  # Code copied from:
  # http://alexleone.blogspot.com/2010/01/python-ast-pretty-printer.html

  def _format(node, level=0):
    """Format a subtree."""

    if isinstance(node, ast.AST):
      fields = [(a, _format(b, level)) for a, b in ast.iter_fields(node)]
      if include_attributes and node._attributes:
        fields.extend([(a, _format(getattr(node, a), level))
                       for a in node._attributes])
      return "".join([
          node.__class__.__name__,
          "(",
          ", ".join(("%s=%s" % field for field in fields)
                    if annotate_fields else
                    (b for a, b in fields)),
          ")"])
    elif isinstance(node, list):
      lines = ["["]
      lines.extend(indent * (level + 2) + _format(x, level + 2) + ","
                   for x in node)
      if len(lines) > 1:
        lines.append(indent * (level + 1) + "]")
      else:
        lines[-1] += "]"
      return "\n".join(lines)
    return repr(node)

  if not isinstance(node, ast.AST):
    raise TypeError(f"expected AST, got {node.__class__!r}")
  return _format(node)



from typing import Any

from .base import BaseStrategy

class LdifProducerStrategy(BaseStrategy):
    sync: bool
    no_real_dsa: bool
    pooled: bool
    can_stream: bool
    line_separator: Any
    all_base64: bool
    stream: Any
    order: Any
    def __init__(self, ldap_connection) -> None: ...
    def receiving(self) -> None: ...
    def send(self, message_type, request, controls: Any | None = ...): ...
    def post_send_single_response(self, message_id): ...
    def post_send_search(self, message_id) -> None: ...
    def accumulate_stream(self, fragment) -> None: ...
    def get_stream(self): ...
    def set_stream(self, value) -> None: ...



from typing import Any, Text

from .models import Response
from .sessions import _Data, _Params

def request(
    method: Text | bytes,
    url: Text | bytes,
    params: _Params | None = ...,
    data: Any | None = ...,
    headers: Any | None = ...,
    cookies: Any | None = ...,
    files: Any | None = ...,
    auth: Any | None = ...,
    timeout: Any | None = ...,
    allow_redirects: bool = ...,
    proxies: Any | None = ...,
    hooks: Any | None = ...,
    stream: Any | None = ...,
    verify: Any | None = ...,
    cert: Any | None = ...,
    json: Any | None = ...,
) -> Response: ...
def get(
    url: Text | bytes,
    params: _Params | None = ...,
    data: Any | None = ...,
    headers: Any | None = ...,
    cookies: Any | None = ...,
    files: Any | None = ...,
    auth: Any | None = ...,
    timeout: Any | None = ...,
    allow_redirects: bool = ...,
    proxies: Any | None = ...,
    hooks: Any | None = ...,
    stream: Any | None = ...,
    verify: Any | None = ...,
    cert: Any | None = ...,
    json: Any | None = ...,
) -> Response: ...
def options(
    url: Text | bytes,
    params: _Params | None = ...,
    data: Any | None = ...,
    headers: Any | None = ...,
    cookies: Any | None = ...,
    files: Any | None = ...,
    auth: Any | None = ...,
    timeout: Any | None = ...,
    allow_redirects: bool = ...,
    proxies: Any | None = ...,
    hooks: Any | None = ...,
    stream: Any | None = ...,
    verify: Any | None = ...,
    cert: Any | None = ...,
    json: Any | None = ...,
) -> Response: ...
def head(
    url: Text | bytes,
    params: _Params | None = ...,
    data: Any | None = ...,
    headers: Any | None = ...,
    cookies: Any | None = ...,
    files: Any | None = ...,
    auth: Any | None = ...,
    timeout: Any | None = ...,
    allow_redirects: bool = ...,
    proxies: Any | None = ...,
    hooks: Any | None = ...,
    stream: Any | None = ...,
    verify: Any | None = ...,
    cert: Any | None = ...,
    json: Any | None = ...,
) -> Response: ...
def post(
    url: Text | bytes,
    data: _Data = ...,
    json: Any | None = ...,
    params: _Params | None = ...,
    headers: Any | None = ...,
    cookies: Any | None = ...,
    files: Any | None = ...,
    auth: Any | None = ...,
    timeout: Any | None = ...,
    allow_redirects: bool = ...,
    proxies: Any | None = ...,
    hooks: Any | None = ...,
    stream: Any | None = ...,
    verify: Any | None = ...,
    cert: Any | None = ...,
) -> Response: ...
def put(
    url: Text | bytes,
    data: _Data = ...,
    params: _Params | None = ...,
    headers: Any | None = ...,
    cookies: Any | None = ...,
    files: Any | None = ...,
    auth: Any | None = ...,
    timeout: Any | None = ...,
    allow_redirects: bool = ...,
    proxies: Any | None = ...,
    hooks: Any | None = ...,
    stream: Any | None = ...,
    verify: Any | None = ...,
    cert: Any | None = ...,
    json: Any | None = ...,
) -> Response: ...
def patch(
    url: Text | bytes,
    data: _Data = ...,
    params: _Params | None = ...,
    headers: Any | None = ...,
    cookies: Any | None = ...,
    files: Any | None = ...,
    auth: Any | None = ...,
    timeout: Any | None = ...,
    allow_redirects: bool = ...,
    proxies: Any | None = ...,
    hooks: Any | None = ...,
    stream: Any | None = ...,
    verify: Any | None = ...,
    cert: Any | None = ...,
    json: Any | None = ...,
) -> Response: ...
def delete(
    url: Text | bytes,
    params: _Params | None = ...,
    data: Any | None = ...,
    headers: Any | None = ...,
    cookies: Any | None = ...,
    files: Any | None = ...,
    auth: Any | None = ...,
    timeout: Any | None = ...,
    allow_redirects: bool = ...,
    proxies: Any | None = ...,
    hooks: Any | None = ...,
    stream: Any | None = ...,
    verify: Any | None = ...,
    cert: Any | None = ...,
    json: Any | None = ...,
) -> Response: ...
